"""
VDOStatus - command ``/usr/bin/vdo status``
===========================================

Module for parsing the output of command ``vdo status``. The bulk of the
content is split on the colon and keys are kept as is.

"""

from __future__ import division
from insights import parser
from insights import YAMLParser
from insights.specs import Specs


@parser(Specs.vdo_status)
class VDOStatus(YAMLParser):
    """
    Class for parsing ``vdo status`` command output.

    Typical output of command ``vdo status`` looks like::

        VDO status:
          Date: '2019-07-27 04:40:40-04:00'
          Node: rdma-qe-04.lab.bos.redhat.com
        Kernel module:
          Loaded: true
          Name: kvdo
          Version information:
            kvdo version: 6.1.0.153
        Configuration:
          File: /etc/vdoconf.yml
          Last modified: '2019-07-26 05:07:48'
        VDOs:
          vdo1:
            Acknowledgement threads: 1
            Activate: enabled
            Device mapper status: 0 8370216 vdo /dev/sda5 albserver online cpu=2,bio=4,ack=1,bioRotationInterval=64
            Physical size: 7G
            Slab size: 2G
            Storage device: /dev/sda5
            VDO statistics:
              /dev/mapper/vdo1:
                overhead blocks used: 787140
                physical blocks: 1835008
                data blocks used: 0
          vdo2:
            Acknowledgement threads: 1
            Activate: enabled
            Device mapper status: 0 8370212 vdo /dev/sda6 albserver online cpu=2,bio=4,ack=1,bioRotationInterval=64
            VDO statistics:
              /dev/mapper/vdo1:
                1K-blocks: 7340032

    Examples:
        >>> type(vdo)
        <class 'insights.parsers.vdo_status.VDOStatus'>
        >>> vdo['Kernel module']['Name']
        'kvdo'
        >>> vdo['Configuration']['File']
        '/etc/vdoconf.yml'
        >>> vdo['VDOs']['vdo1']['Activate']
        'enabled'
        >>> vdo['VDOs']['vdo1']['VDO statistics']['/dev/mapper/vdo1']['1K-blocks']
        7340032
        >>> vdo['VDO status']
        {'Date': '2019-07-24 20:48:16-04:00', 'Node': 'dell-m620-10.rhts.gsslab.pek2.redhat.com'}
        >>> vdo['VDOs']['vdo2']['Acknowledgement threads']
        1
        >>> vdo.get_slab_size_of_vol('vdo1')
        '2G'
        >>> vdo.volumns
        ['vdo1', 'vdo2']
        >>> vdo.get_physical_blocks_of_vol('vdo1')
        1835008
        >>> vdo.get_physical_used_of_vol('vdo1')
        0
        >>> vdo.get_physical_free_of_vol('vdo1')
        1047868
        >>> vdo.get_logical_used_of_vol('vdo1')
        0
        >>> vdo.get_overhead_used_of_vol('vdo1')
        787140

    Raises:
        ParseException: When input content is not available to parse

    Attributes:
        data (dict): the result parsed of 'vdo status'
        volumns (list): The list the vdo volumns involved

    """

    def __get_dev_mapper__(self, vol):
        """
        Device mapper path of a specified vdo

        Args:
            vol: String of vdo volumne name

        Returns:
            dict: Device mapper information of a specified volumne

        Raises:
            KeyError: If KEYs doesn't exist
        """

        try:
            dm_path = ('/dev/mapper/%s' % vol)
            return self.data['VDOs'][vol]['VDO statistics'][dm_path]
        except:
            err_path = "['VDOs'][%s]['VDO statistics'][%s]" % (vol, dm_path)
            raise KeyError('No key(s) named: %s' % err_path)

    @property
    def volumns(self):
        """
        The volumns appeared in vdo status

        Returns:
            list: vdo volumns
        """
        return sorted(self.data['VDOs'].keys()) if 'VDOs' in self.data else []

    def get_slab_size_of_vol(self, vol):
        """
        The slab size of a specified volumne

        Args:
            vol (str): The vdo volumne name specified

        Returns:
            str: Slab size of specified vdo volumne
        """
        return self.data['VDOs'][vol]['Slab size']

    def get_physical_blocks_of_vol(self, vol):
        """
        The physical blocks of a specified volumne

        Args:
            vol (str): The vdo volumne name specified

        Returns:
            int: physical blocks size
        """
        dm = self.__get_dev_mapper__(vol)
        return dm['physical blocks']

    def get_physical_used_of_vol(self, vol):
        """
        The physical used blocks of a specified volumn

        Args:
            vol (str): The vdo volumne name specified

        Returns:
            int: Returns size of physical blocks used
        """
        dm = self.__get_dev_mapper__(vol)
        return dm['data blocks used']

    def get_overhead_used_of_vol(self, vol):
        """
        The overhead used blocks of a specified volumne

        Args:
            vol (str): The vdo volumne name specified

        Returns:
            int: Returns size of overhead blocks used
        """
        dm = self.__get_dev_mapper__(vol)
        return dm['overhead blocks used']

    def get_logical_blocks_of_vol(self, vol):
        """
        The logical blocks of a specified volumne

        Args:
            vol (str): The vdo volumne name specified

        Returns:
            int: Returns size of logical blocks
        """
        dm = self.__get_dev_mapper__(vol)
        return dm['logical blocks']

    def get_logical_used_of_vol(self, vol):
        """
        The logical used blocks of a specified volumne

        Args:
            vol (str): The vdo volumne name specified

        Returns:
            int: Returns size of logical blocks used
        """
        dm = self.__get_dev_mapper__(vol)
        return dm['logical blocks used']

    def get_logical_free_of_vol(self, vol):
        """
        The logical free blocks of a specified volumne

        Args:
            vol (str): The vdo volumne name specified

        Returns:
            int: Returns size of logical free
        """
        return (self.get_logical_blocks_of_vol(vol) -
                self.get_logical_used_of_vol(vol))

    def get_physical_free_of_vol(self, vol):
        """
        The physical free blocks of a specified volumne

        Args:
            vol (str): The vdo volumne name specified

        Returns:
            int: Returns size of physical free
        """
        return (self.get_physical_blocks_of_vol(vol) -
                self.get_overhead_used_of_vol(vol) -
                self.get_physical_used_of_vol(vol))



"""
Combined NFS exports
====================

The NFS exports files are normally available to rules from both a single
NFSExports object and zero or more NFSExportsD objects.  This combiner turns
those into one set of data.

Examples:
    >>> type(all_nfs)
    <class 'insights.combiners.nfs_exports.AllNFSExports'>
    >>> all_nfs.files  # List of files exporting NFS shares
    ['/etc/exports', '/etc/exports.d/mnt.exports']
    >>> '/home/example' in all_nfs.exports  # All exports stored by path
    True
    >>> sorted(all_nfs.exports['/home/example'].keys())  # Each path is a dictionary of host specs and flags.
    ['@group', 'ins1.example.com', 'ins2.example.com']
    >>> all_nfs.exports['/home/example']['ins2.example.com']  # Each host contains a list of flags.
    ['rw', 'sync', 'no_root_squash']
    >>> '/home/example' in all_nfs.ignored_exports  # Ignored exports are remembered within one file
    True
    >>> list(all_nfs.ignored_exports['/home/example'].keys())  # Each ignored export is then stored by source file...
    ['/etc/exports']
    >>> list(all_nfs.ignored_exports['/home/example']['/etc/exports'].keys())  # ... and then by host spec...
    ['ins2.example.com']
    >>> all_nfs.ignored_exports['/home/example']['/etc/exports']['ins2.example.com']  # ... holding the values that were duplicated
    ['rw', 'sync', 'no_root_squash']
    >>> '/home/insights/shared/rw'  in all_nfs.ignored_exports  # Ignored exports are remembered across files
    True

"""

from insights.core.plugins import combiner
from insights.parsers.nfs_exports import NFSExports, NFSExportsD

import collections


@combiner(NFSExports, optional=[NFSExportsD])
class AllNFSExports(object):
    """
    Combiner for accessing all the NFS export configuration files.

    Exports are allowed to be listed multiple times, with all duplicate
    host after the first causing ``exportfs`` to emit a warning and ignore
    the host.  So we combine the raw lines and ignored exports into structures
    listing the source file for each

    Attributes:
        files (list): the list of source files that contained NFS export
            definitions.

        exports (dict of dicts): the NFS exports stored by export path, with
            each path storing a dictionary of host flag lists.

        ignored_exports (dict): A dictionary of exported paths that have
            host definitions that conflicted with a previous definition,
            stored by export path and then path of the file that defined it.

        raw_lines (dict of dicts): A dictionary of raw lines that define each
            exported path, with the lines stored by defining file.

    """
    def __init__(self, nfsexports, nfsexportsd):
        self.files = []
        self.exports = {}
        self.ignored_exports = {}
        self.raw_lines = {}

        sources = [nfsexports]
        # Make sure exports are stored in the order they're parsed -
        # alphabetically by file name.  Ignore it if nfsexportsd isn't valid.
        if isinstance(nfsexportsd, collections.Iterable):
            sources.extend(sorted(nfsexportsd, key=lambda f: f.file_path))

        def add_paths_to_dict(src_path, src_dict, dest_dict):
            # Because ignored_exports and raw_lines are stored by path, we
            # keep that and add the paths in them piecewise, stored by the
            # path of this source file.
            for path, value in src_dict.items():
                if path not in dest_dict:
                    dest_dict[path] = {src_path: value}
                else:
                    dest_dict[path][src_path] = value

        for source in sources:
            self.files.append(source.file_path)
            # Add all raw lines from the source to raw lines.
            add_paths_to_dict(source.file_path, source.raw_lines, self.raw_lines)
            # Likewise all the ignored exports from this file.
            add_paths_to_dict(source.file_path, source.ignored_exports, self.ignored_exports)
            # For exports though we have to preserve existing host definitions
            # and ignore repeated host specs.
            for path, hosts in source.data.items():
                if path in self.exports:
                    # Check whether the host specification has been listed
                    # more than once across different files:
                    for host, flags in hosts.items():
                        if host in self.exports[path]:
                            # Add this to the ignored_exports list.  But
                            # because we know that this path-host tuple isn't
                            # duplicated in this file (it's already in
                            # ignored_exports if it does), then we can assume
                            # that the file isn't in the common ignored_exports
                            # already.
                            if path in self.ignored_exports:
                                self.ignored_exports[path][host] = flags
                            else:
                                self.ignored_exports[path] = {host: flags}
                        else:
                            # A new host spec - add as is.
                            self.exports[path][host] = flags
                else:
                    # No exports for this path so far, add them all
                    self.exports[path] = hosts

        super(AllNFSExports, self).__init__()



"""
Simple arithmetic with usual operator precedence and associativity. It allows
addition, subtraction, multiplication, division, and grouping with parentheses.
"""
from insights.parsr import (EOF, Forward, InSet, LeftParen, Many, Number,
        RightParen, WS)


def evaluate(e):
    return Top(e)[0]


def op(args):
    ans, rest = args
    for op, arg in rest:
        if op == "+":
            ans += arg
        elif op == "-":
            ans -= arg
        elif op == "*":
            ans *= arg
        elif op == "/":
            ans /= arg
    return ans


# high precedence operations
HighOps = InSet("*/")

# low precedence operations
LowOps = InSet("+-")

# Operator precedence is handled by having different declarations for each
# prededence level. expr handles low level operations, term handles high level
# operations, and factor handles simple numbers or subexpressions between
# parentheses. Since the first element in expr is term and the first element in
# term is factor, factors are evaluated first, then terms, and then exprs.
#
# The Top declaration ensures that the entire input is valid with no extra data
# at the end.

# We have to declare expr before its definition since it's used recursively.
expr = Forward() % "expr forward"


# A factor is a simple number or a subexpression between parentheses
factor = WS >> (Number % "Number" | (LeftParen >> expr << RightParen)) << WS

# A term handles strings of multiplication and division. As written, it would
# convert "1 + 2 - 3 + 4" into [1, [['+', 2], ['-', 3], ['+', 4]]]. The first
# element in the outer list is the initial factor. The second element of the
# outer list is another list, which is the result of the Many. The Many's list
# contains several two-element lists generated from each match of
# (HighOps + factor). We pass the entire structure into the op function with
# map.
term = (factor + Many(HighOps + factor)).map(op) % "term"

# expr has the same form as term.
expr <= (term + Many(LowOps + term)).map(op) % "expr"

# Top returns [result, None] on success and raises an Exception on failure.
Top = (expr + EOF) % "Top"



"""
AWSInstanceType
===============

This parser simply reads the output of command
``curl http://169.254.169.254/latest/meta-data/instance-type``,
which is used to check the type of the AWS instance on the host.

"""

from insights.parsers import SkipException, ParseException
from insights import parser, CommandParser
from insights.specs import Specs


@parser(Specs.aws_instance_type)
class AWSInstanceType(CommandParser):
    """
    Class for parsing the AWS Instance type returned by command
    ``curl http://169.254.169.254/latest/meta-data/instance-type``

    Typical output of this command is::

        r3.xlarge

    Raises:
        SkipException: When content is empty or no parse-able content.
        ParseException: When type cannot be recognized.

    Attributes:
        type (str): The name of AWS instance type in all uppercase letters. E.g. R3, R4, R5, or X1.
        raw (str): The fully type string returned by the ``curl`` command.

    Examples:
        >>> aws_inst.type
        'R3'
        >>> aws_inst.raw
        'r3.xlarge'
    """

    def parse_content(self, content):
        if (not content or len(content) > 1 or 'curl: ' in content[0]):
            raise SkipException()

        self.raw = self.type = None
        if '.' in content[0]:
            self.raw = content[0].strip()
            self.type = self.raw.split('.')[0].upper()

        if not self.type:
            raise ParseException('Unrecognized type: "{0}"', content[0])

    def __repr__(self):
        return "<aws_type: {t}, raw: {r}>".format(t=self.type, r=self.raw)



"""
Utility functions
"""
from __future__ import absolute_import
import socket
import os
import logging
import uuid
import datetime
import shlex
import re
import stat
import sys
from subprocess import Popen, PIPE, STDOUT
from six.moves.configparser import RawConfigParser

from .constants import InsightsConstants as constants

logger = logging.getLogger(__name__)


def determine_hostname(display_name=None):
    """
    Find fqdn if we can
    """
    if display_name:
        # if display_name is provided, just return the given name
        return display_name
    else:
        socket_gethostname = socket.gethostname()
        socket_fqdn = socket.getfqdn()

        try:
            socket_ex = socket.gethostbyname_ex(socket_gethostname)[0]
        except (LookupError, socket.gaierror):
            socket_ex = ''

        gethostname_len = len(socket_gethostname)
        fqdn_len = len(socket_fqdn)
        ex_len = len(socket_ex)

        if fqdn_len > gethostname_len or ex_len > gethostname_len:
            if "localhost" not in socket_ex and len(socket_ex):
                return socket_ex
            if "localhost" not in socket_fqdn:
                return socket_fqdn

        return socket_gethostname


def get_time():
    return datetime.datetime.isoformat(datetime.datetime.now())


def write_registered_file():
    delete_unregistered_file()
    for f in constants.registered_files:
        if os.path.lexists(f):
            if os.path.islink(f):
                # kill symlinks and regenerate
                os.remove(f)
                write_to_disk(f)
        else:
            write_to_disk(f)


def write_unregistered_file(date=None):
    """
    Write .unregistered out to disk
    """
    delete_registered_file()
    if date is None:
        date = get_time()
    for f in constants.unregistered_files:
        if os.path.lexists(f):
            if os.path.islink(f):
                # kill symlinks and regenerate
                os.remove(f)
                write_to_disk(f, content=str(date))
        else:
            write_to_disk(f, content=str(date))


def delete_registered_file():
    for f in constants.registered_files:
        write_to_disk(f, delete=True)


def delete_unregistered_file():
    for f in constants.unregistered_files:
        write_to_disk(f, delete=True)


def write_to_disk(filename, delete=False, content=get_time()):
    """
    Write filename out to disk
    """
    if not os.path.exists(os.path.dirname(filename)):
        return
    if delete:
        if os.path.lexists(filename):
            os.remove(filename)
    else:
        with open(filename, 'wb') as f:
            f.write(content.encode('utf-8'))


def generate_machine_id(new=False,
                        destination_file=constants.machine_id_file):
    """
    Generate a machine-id if /etc/insights-client/machine-id does not exist
    """
    machine_id = None
    machine_id_file = None
    logging_name = 'machine-id'

    if os.path.isfile(destination_file) and not new:
        logger.debug('Found %s', destination_file)
        with open(destination_file, 'r') as machine_id_file:
            machine_id = machine_id_file.read()
    else:
        logger.debug('Could not find %s file, creating', logging_name)
        machine_id = str(uuid.uuid4())
        logger.debug("Creating %s", destination_file)
        write_to_disk(destination_file, content=machine_id)

    return str(machine_id).strip()


def _expand_paths(path):
    """
    Expand wildcarded paths
    """
    dir_name = os.path.dirname(path)
    paths = []
    logger.debug("Attempting to expand %s", path)
    if os.path.isdir(dir_name):
        files = os.listdir(dir_name)
        match = os.path.basename(path)
        for file_path in files:
            if re.match(match, file_path):
                expanded_path = os.path.join(dir_name, file_path)
                paths.append(expanded_path)
        logger.debug("Expanded paths %s", paths)
        return paths
    else:
        logger.debug("Could not expand %s", path)


def validate_remove_file(remove_file):
    """
    Validate the remove file
    """
    if not os.path.isfile(remove_file):
        logger.warn("WARN: Remove file does not exist")
        return False
    # Make sure permissions are 600
    mode = stat.S_IMODE(os.stat(remove_file).st_mode)
    if not mode == 0o600:
        logger.error("ERROR: Invalid remove file permissions"
                     "Expected 0600 got %s" % oct(mode))
        return False
    else:
        logger.debug("Correct file permissions")

    if os.path.isfile(remove_file):
        parsedconfig = RawConfigParser()
        parsedconfig.read(remove_file)
        rm_conf = {}
        for item, value in parsedconfig.items('remove'):
            rm_conf[item] = value.strip().split(',')
        # Using print here as this could contain sensitive information
        logger.debug("Remove file parsed contents")
        logger.debug(rm_conf)
    logger.info("JSON parsed correctly")
    return True


def write_data_to_file(data, filepath):
    '''
    Write data to file
    '''
    try:
        os.makedirs(os.path.dirname(filepath), 0o700)
    except OSError:
        pass

    write_to_disk(filepath, content=data)


def magic_plan_b(filename):
    '''
    Use this in instances where
    python-magic is MIA and can't be installed
    for whatever reason
    '''
    cmd = shlex.split('file --mime-type --mime-encoding ' + filename)
    stdout, stderr = Popen(cmd, stdout=PIPE).communicate()
    stdout = stdout.decode("utf-8")
    mime_str = stdout.split(filename + ': ')[1].strip()
    return mime_str


def run_command_get_output(cmd):
    proc = Popen(shlex.split(cmd),
                 stdout=PIPE, stderr=STDOUT)
    stdout, stderr = proc.communicate()

    return {
        'status': proc.returncode,
        'output': stdout.decode('utf-8', 'ignore')
    }


def modify_config_file(updates):
    '''
    Update the config file with certain things
    '''
    cmd = '/bin/sed '
    for key in updates:
        cmd = cmd + '-e \'s/^#*{key}.*=.*$/{key}={value}/\' '.format(key=key, value=updates[key])
    cmd = cmd + constants.default_conf_file
    status = run_command_get_output(cmd)
    write_to_disk(constants.default_conf_file, content=status['output'])


def get_version_info():
    '''
    Get the insights client and core versions for archival
    '''
    from insights.client import InsightsClient

    cmd = 'rpm -q --qf "%{VERSION}-%{RELEASE}" insights-client'
    version_info = {}
    version_info['core_version'] = InsightsClient().version()
    version_info['client_version'] = run_command_get_output(cmd)['output']

    return version_info


def print_egg_versions():
    '''
    Log all available eggs' versions
    '''
    versions = get_version_info()
    logger.debug('Client version: %s', versions['client_version'])
    logger.debug('Core version: %s', versions['core_version'])
    logger.debug('All egg versions:')
    eggs = [
        os.getenv('EGG'),
        '/var/lib/insights/newest.egg',
        '/var/lib/insights/last_stable.egg',
        '/etc/insights-client/rpm.egg',
    ]
    if not sys.executable:
        logger.debug('Python executable not found.')
        return

    for egg in eggs:
        if egg is None:
            logger.debug('ENV egg not defined.')
            continue
        if not os.path.exists(egg):
            logger.debug('%s not found.', egg)
            continue
        try:
            proc = Popen([sys.executable, '-c', 'from insights.client import InsightsClient; print(InsightsClient().version())'],
                         env={'PYTHONPATH': egg, 'PATH': os.getenv('PATH')}, stdout=PIPE, stderr=STDOUT)
        except OSError:
            logger.debug('Could not start python.')
            return
        stdout, stderr = proc.communicate()
        version = stdout.decode('utf-8', 'ignore').strip()
        logger.debug('%s: %s', egg, version)



from insights.parsers import samba
from insights.tests import context_wrap
from doctest import testmod

SAMBA_CONFIG_DOCUMENTATION = '''
# This is the main Samba configuration file. You should read the
# smb.conf(5) manual page in order to understand the options listed
#...
#======================= Global Settings =====================================

[global]
    workgroup = MYGROUP
    server string = Samba Server Version %v
    max log size = 50

[homes]
    comment = Home Directories
    browseable = no
    writable = yes
;   valid users = %S
;   valid users = MYDOMAIN\%S

[printers]
    comment = All Printers
    path = /var/spool/samba
    browseable = no
    guest ok = no
    writable = no
    printable = yes

# A publicly accessible directory, but read only, except for people in
# the "staff" group
[public]
   comment = Public Stuff
   path = /home/samba
   public = yes
   writable = yes
   printable = no
   write list = +staff
'''


def test_documentation():
    failed, total = testmod(samba, globs={
        'conf': samba.SambaConfig(context_wrap(SAMBA_CONFIG_DOCUMENTATION)),
    })
    assert failed == 0


SAMBA_CONFIG = """
# This is the main Samba configuration file. You should read the
# smb.conf(5) manual page in order to understand the options listed
#...
#======================= Global Settings =====================================

# By running `testparm`, it is apparent that options outside the [global]
# section before the [global] section are treated as if they were in the
# [global] section.
# `testparm` is a program from the samba package.

this option should be in global = yes

[global]

#...
# Hosts Allow/Hosts Deny lets you restrict who can connect, and you can
# specifiy it as a per share option as well
#
    workgroup = MYGROUP
    server string = Samba Server Version %v

;   netbios name = MYSERVER

;   interfaces = lo eth0 192.168.12.2/24 192.168.13.2/24
;   hosts allow = 127. 192.168.12. 192.168.13.

# --------------------------- Logging Options -----------------------------
#
# Log File let you specify where to put logs and how to split them up.
#
# Max Log Size let you specify the max size log files should reach

    # logs split per machine
    log file = /var/log/samba/log.%m
    # max 50KB per log file, then rotate
    max log size = 50

# ----------------------- Standalone Server Options ------------------------
#
# Scurity can be set to user, share(deprecated) or server(deprecated)
#
# Backend to store user information in. New installations should
# use either tdbsam or ldapsam. smbpasswd is available for backwards
# compatibility. tdbsam requires no further configuration.

    security = user
    passdb backend = tdbsam

#...
# --------------------------- Printing Options -----------------------------
#
# Load Printers let you load automatically the list of printers rather
# than setting them up individually
#
# Cups Options let you pass the cups libs custom options, setting it to raw
# for example will let you use drivers on your Windows clients
#
# Printcap Name let you specify an alternative printcap file
#
# You can choose a non default printing system using the Printing option

    load printers = yes
    cups options = raw

;   printcap name = /etc/printcap
    #obtain list of printers automatically on SystemV
;   printcap name = lpstat
;   printing = cups

# --------------------------- Filesystem Options ---------------------------
#
# The following options can be uncommented if the filesystem supports
# Extended Attributes and they are enabled (usually by the mount option
# user_xattr). Thess options will let the admin store the DOS attributes
# in an EA and make samba not mess with the permission bits.
#
# Note: these options can also be set just per share, setting them in global
# makes them the default for all shares

;   map archive = no
;   map hidden = no
;   map read only = no
;   map system = no
;   store dos attributes = yes


#============================ Share Definitions ==============================

[homes]
    comment = Home Directories
    browseable = no
    writable = yes
;   valid users = %S
;   valid users = MYDOMAIN\%S

[printers]
    comment = All Printers
    path = /var/spool/samba
    browseable = no
    guest ok = no
    writable = no
    printable = yes

# Un-comment the following and create the netlogon directory for Domain Logons
;   [netlogon]
;   comment = Network Logon Service
;   path = /var/lib/samba/netlogon
;   guest ok = yes
;   writable = no
;   share modes = no


# Un-comment the following to provide a specific roving profile share
# the default is to use the user's home directory
;   [Profiles]
;   path = /var/lib/samba/profiles
;   browseable = no
;   guest ok = yes


# A publicly accessible directory, but read only, except for people in
# the "staff" group
;   [public]
;   comment = Public Stuff
;   path = /home/samba
;   public = yes
;   writable = yes
;   printable = no
;   write list = +staff

[ GlObAl  ]

# Samba also automatically treats non-lowercase section names as lowercase and strips whitespace.
# This behavior can be checked with `testparm` again.
this option should also be in global = true

[ GlObAl  ]

# This tests specifically that two same-named sections are automatically merged in case the
# RawConfigParser's behavior ever changes.
this another option should also be in global = 1
"""


def test_match():
    config = samba.SambaConfig(context_wrap(SAMBA_CONFIG))

    assert config.get('global', 'this option should be in global') == 'yes'
    assert config.get('global', 'this option should also be in global') == 'true'
    assert config.get('global', 'this another option should also be in global') == '1'
    assert config.get('global', 'workgroup') == 'MYGROUP'
    assert config.get('global', 'workgroup') == 'MYGROUP'
    assert config.get('global', 'server string') == 'Samba Server Version %v'
    assert not config.has_option('global', 'netbios name')
    assert config.get('global', 'log file') == '/var/log/samba/log.%m'
    assert config.get('global', 'max log size') == '50'

    assert config.get('global', 'security') == 'user'
    assert config.get('global', 'passdb backend') == 'tdbsam'

    assert config.get('global', 'load printers') == 'yes'
    assert config.get('global', 'cups options') == 'raw'

    assert not config.has_option('global', 'printcap name')

    assert config.get('homes', 'comment') == 'Home Directories'
    assert config.get('homes', 'browseable') == 'no'
    assert config.get('homes', 'writable') == 'yes'
    assert not config.has_option('homes', 'valid users')

    assert config.get('printers', 'comment') == 'All Printers'
    assert config.get('printers', 'path') == '/var/spool/samba'
    assert config.get('printers', 'browseable') == 'no'
    assert config.get('printers', 'guest ok') == 'no'
    assert config.get('printers', 'writable') == 'no'
    assert config.get('printers', 'printable') == 'yes'

    assert 'netlogin' not in config
    assert 'Profiles' not in config
    assert 'public' not in config



"""
IBM DB2 Sever details
=====================

Module for the processing of output from the ``db2licm -l`` command.
"""
from insights.core.plugins import parser
from insights.core import CommandParser
from insights.parsers import ParseException, get_active_lines
from insights.specs import Specs


@parser(Specs.db2licm_l)
class DB2Info(CommandParser, dict):
    """
    This parser processes the output of the command `db2licm_l` and provides
    the information as a dictionary.

    Sample input::

        Product name:                     "DB2 Enterprise Server Edition"
        License type:                     "CPU Option"
        Expiry date:                      "Permanent"
        Product identifier:               "db2ese"
        Version information:              "9.7"
        Enforcement policy:               "Soft Stop"
        Features:
        DB2 Performance Optimization ESE: "Not licensed"
        DB2 Storage Optimization:         "Not licensed"
        DB2 Advanced Access Control:      "Not licensed"
        IBM Homogeneous Replication ESE:  "Not licensed"

        Product name:                     "DB2 Connect Server"
        Expiry date:                      "Expired"
        Product identifier:               "db2consv"
        Version information:              "9.7"
        Concurrent connect user policy:   "Disabled"
        Enforcement policy:               "Soft Stop"

    Example:

        >>> list(parser_result.keys())
        ['DB2 Enterprise Server Edition', 'DB2 Connect Server']
        >>> parser_result['DB2 Enterprise Server Edition']["Version information"]
        '9.7'

    Override the base class parse_content to parse the output of the '''db2licm -l'''  command.
    Information that is stored in the object is made available to the rule plugins.


    Raises:
        ParseException: raised if data is not parsable.
    """

    def parse_content(self, content):

        # name = None
        body = {}

        # Input data is available in text file. Reading each line in file and parsing it to a dictionary.
        for line in get_active_lines(content):
            if ':' in line:
                key, val = [i.strip() for i in line.strip().split(":", 1)]
                if key == "Features":
                    continue
            else:
                raise ParseException("Unable to parse db2licm info: {0}".format(content))

            if key == "Product name":
                body = {}
                self[val] = body
            else:
                body[key] = val

        if not self:
            # If no data is obtained in the command execution then throw an exception instead of returning an empty
            # object.  Rules depending solely on this parser will not be invoked, so they don't have to
            # explicitly check for invalid data.
            raise ParseException("Unable to parse db2licm info: {0}".format(content))



"""
HammerPing - command ``/usr/bin/hammer ping``
=============================================

The hammer ping parser reads the output of ``hammer ping`` and turns it into
a dictionary of service statuses in the ``status_of_service`` property and
the corresponding response in the ``response_of_service`` property.  The list
of services by name in the order they were discovered is given in the
``service_list`` property.

Sample output of ``hammer ping``::

    candlepin:
        Status:          FAIL
        Server Response: Message: 404 Resource Not Found
    elasticsearch:
        Status:          ok
        Server Response: Duration: 35ms
    foreman_tasks:
        Status:          ok
        Server Response: Duration: 1ms

Examples:

    >>> hammer = shared[HammerPing]
    >>> 'unknown_service' in hammer.service_list
    False
    >>> hammer.status_of_service['candlepin']
    'FAIL'
    >>> hammer.response_of_service['candlepin']
    'Message: 404 Resource Not Found'
    >>> hammer.are_all_ok
    False
    >>> hammer.services_of_status('OK')
    ['elasticsearch', 'foreman_tasks']
"""
from .. import parser, CommandParser
from insights.specs import Specs


@parser(Specs.hammer_ping)
class HammerPing(CommandParser):
    """
    Read the ``hammer ping`` status and convert it to dictionaries of
    status and response information.

    Attributes:
        status_of_service (dict): The status of each service, converted to lower case
        response_of_service (dict): The response of each service, as is
        service_list (list): The list of service names in order
        errors (list): Any error messages encountered during parsing
        are_all_ok (bool): Are all services read correctly and are they all in 'ok' state?

    """
    def services_of_status(self, status='ok'):
        """
        List of the services in the given status.

        Arguments:
            status (str): the status code to search for, defaulting to 'ok'.
                The value is converted to lower case.

        Returns: List of service names having that status.
        """
        status = status.lower()
        return [svc for svc in self.service_list
                if self.status_of_service[svc] == status]

    def parse_content(self, content):
        self.status_of_service = {}
        self.response_of_service = {}
        self.service_list = []
        self.errors = []
        self.are_all_ok = False

        # The first line should be a service
        if len(content[0].split()) > 1:
            self.errors.append(content[0])
        # Each service occupies 3 lines
        elif len(content) % 3 == 0:
            lines = iter(content)
            for service_no in range(0, len(content) // 3):
                service_line = next(lines).strip()
                status_line = next(lines).strip()
                response_line = next(lines).strip()

                service = service_line.split(':')[0]
                status = status_line.split(':', 1)[-1].strip().lower()
                response = response_line.split(':', 1)[-1].strip()

                self.service_list.append(service)
                self.status_of_service[service] = status
                self.response_of_service[service] = response
        else:
            self.errors.extend(content)

        self.are_all_ok = (
            not self.errors and
            all(status == 'ok' for status in self.status_of_service.values())
        )



"""
Yum - Commands
==============
Parsers for ``yum`` commands.

This module contains the classes that parse the output of the commands
`yum -C repolist`.

YumRepoList - command ``yum -C repolist``
-----------------------------------------
"""

from insights import parser, CommandParser
from insights.parsers import SkipException, parse_fixed_table
from insights.specs import Specs

eus = [
    '5.0.z',
    '5.1.z',
    '5.2.z',
    '5.3.ll',
    '5.3.z',
    '5.4.z',
    '5.6.ll',
    '5.6.z',
    '5.9.ll',
    '5.9.z'
    '6.0.z',
    '6.1.z',
    '6.2.aus',
    '6.2.z',
    '6.3.z',
    '6.4.aus',
    '6.4.z',
    '6.5.aus',
    '6.5.z',
    '6.6.aus',
    '6.6.z',
    '6.7.z'
]


@parser(Specs.yum_repolist)
class YumRepoList(CommandParser):
    """
    Class for parsing the output of `yum -C repolist` command.

    Typical output of the command is::

        Loaded plugins: langpacks, product-id, search-disabled-repos, subscription-manager
        repo id                                             repo name                                                                                                    status
        rhel-7-server-e4s-rpms/x86_64                       Red Hat Enterprise Linux 7 Server - Update Services for SAP Solutions (RPMs)                                 12,250
        !rhel-ha-for-rhel-7-server-e4s-rpms/x86_64          Red Hat Enterprise Linux High Availability (for RHEL 7 Server) Update Services for SAP Solutions (RPMs)         272
        *rhel-sap-hana-for-rhel-7-server-e4s-rpms/x86_64    RHEL for SAP HANA (for RHEL 7 Server) Update Services for SAP Solutions (RPMs)                                   21
        repolist: 12,768

    Or sometimes it just outputs repo id and status::

        Loaded plugins: package_upload, product-id, search-disabled-repos, security, subscription-manager
        repo id                                                                               status
        LME_EPEL_6_x86_64                                                                        26123
        LME_FSMLabs_Timekeeper_timekeeper                                                            2
        LME_HP_-_Software_Delivery_Repository_Firmware_Pack_for_ProLiant_-_6Server_-_Current      1163
        LME_HP_-_Software_Delivery_Repository_Scripting_Took_Kit_-_6Server                          17
        LME_HP_-_Software_Delivery_Repository_Service_Pack_for_ProLiant_-_6Server_-_Current       1915
        LME_HP_-_Software_Delivery_Repository_Smart_Update_Manager_-_6Server                        30
        LME_LME_Custom_Product_Mellanox_OFED                                                       114
        LME_LME_Custom_Product_OMD_RPMS                                                             14
        LME_LME_Custom_Product_RPMs                                                                  5
        LME_LME_Custom_Product_SNOW_Repository                                                       2
        rhel-6-server-optional-rpms                                                            10400+1
        rhel-6-server-rpms                                                                    18256+12
        rhel-6-server-satellite-tools-6.2-rpms                                                      55
        repolist: 58096

    Examples:
        >>> len(repolist)
        3
        >>> 'rhel-7-server-e4s-rpms/x86_64' in repolist.repos
        True
        >>> 'rhel-7-server-e4s-rpms' in repolist.repos
        False
        >>> 'rhel-7-server-e4s-rpms' in repolist.rhel_repos
        True
        >>> repolist['rhel-7-server-e4s-rpms/x86_64']['name']
        'Red Hat Enterprise Linux 7 Server - Update Services for SAP Solutions (RPMs)'
        >>> repolist[0]['name']
        'Red Hat Enterprise Linux 7 Server - Update Services for SAP Solutions (RPMs)'
        >>> repolist['rhel-ha-for-rhel-7-server-e4s-rpms/x86_64']['id']
        '!rhel-ha-for-rhel-7-server-e4s-rpms/x86_64'
        >>> len(repolist_no_reponame)
        13
        >>> len(repolist_no_reponame.rhel_repos)
        3
        >>> 'rhel-6-server-rpms' in repolist_no_reponame.repos
        True
        >>> 'rhel-6-server-optional-rpms' in repolist_no_reponame.rhel_repos
        True
        >>> repolist_no_reponame[0]['id']
        'LME_EPEL_6_x86_64'
        >>> repolist_no_reponame[0].get('name', '')
        ''

    Attributes:
        data (list): list of repos wrapped in dictionaries
        repos (dict): dict of all listed repos where the key is the full repo-id without "!" or "*".
            But you can get it from the value part if needed. For example::

                self.repos = {
                    'rhel-7-server-e4s-rpms/x86_64': {
                        'id': 'rhel-7-server-e4s-rpms/x86_64',
                        'name': 'Red Hat Enterprise Linux 7 Server - Update Services for SAP Solutions (RPMs)',
                        'status': '12,250'
                    },
                    'rhel-ha-for-rhel-7-server-e4s-rpms/x86_64': {
                        'id': '!rhel-ha-for-rhel-7-server-e4s-rpms/x86_64',
                        'name': 'Red Hat Enterprise Linux High Availability (for RHEL 7 Server) Update Services for SAP Solutions (RPMs)',
                        'status': '272'
                    },
                    'rhel-sap-hana-for-rhel-7-server-e4s-rpms/x86_64': {
                        'id': '*rhel-sap-hana-for-rhel-7-server-e4s-rpms/x86_64',
                        'name': 'RHEL for SAP HANA (for RHEL 7 Server) Update Services for SAP Solutions (RPMs)',
                        'status': '21'
                    }
                }

        rhel_repos(list): list of all the rhel repos and the item is just the repo id without server and arch info. For example::

            self.rhel_repos = ['rhel-7-server-e4s-rpms', 'rhel-ha-for-rhel-7-server-e4s-rpms', 'rhel-sap-hana-for-rhel-7-server-e4s-rpms']
    """
    def parse_content(self, content):
        if not content:
            raise SkipException('No repolist.')

        trailing_line_prefix = [
                'repolist:',
                'Uploading Enabled',
                'Loaded plugins:',
                'Unable to upload',
        ]

        self.data = []
        self.repos = {}
        self.data = parse_fixed_table(
                content,
                heading_ignore=['repo id'],
                header_substitute=[('repo id', 'id     '), ('repo name', 'name     ')],
                trailing_ignore=trailing_line_prefix,
                empty_exception=True)

        if not self.data:
            raise SkipException('No repolist.')

        self.repos = dict((d['id'].lstrip('!').lstrip('*'), d) for d in self.data)

    def __len__(self):
        return len(self.data)

    def __getitem__(self, idx):
        if isinstance(idx, int):
            return self.data[idx]
        elif isinstance(idx, str):
            return self.repos[idx]

    def __contains__(self, _id):
        return _id in self.repos

    @property
    def eus(self):
        '''list of the EUS part of each repo'''
        euses = []
        for repo in [r["id"] for r in self.data]:
            if repo.startswith("rhel-") and "server-" in repo:
                _, eus_version = repo.split("server-", 1)
                if eus_version in eus:
                    euses.append(eus_version)
        return euses

    @property
    def rhel_repos(self):
        '''list of RHEL repos/Repo IDs'''
        return [i.split('/')[0]
                for i in self.repos
                if i.startswith('rhel')]



from insights.parsr.query import all_, any_, lt, Entry, from_dict, startswith, endswith


simple_data = {
    "a": 3,
    "b": ["four", "five"],
    "c": [9, 15],
    "d": {"e": 9.7}
}


complex_tree = Entry(name="root",
                    attrs=[1, 2, 3, 4],
                    children=[
                        Entry(name="child", attrs=[1, 1, 2]),
                        Entry(name="child", attrs=[1, 1, 2, 3, 5]),
                        Entry(name="child", attrs=[1, 1, 3, 5, 9]),
                        Entry(name="dog", attrs=["woof"], children=[
                            Entry(name="puppy", attrs=["smol"]),
                            Entry(name="puppy", attrs=["fluffy"]),
                            Entry(name="kitten", attrs=["wut"]),
                        ])
                    ])


def test_from_dict():
    n = from_dict(simple_data)
    assert n
    assert len(n) == 4


def test_values():
    n = from_dict(simple_data)
    assert n["a"].value == 3

    assert n["b"].value == "four five"
    assert n["b"].string_value == "four five"

    assert n["c"].value == "9 15"
    assert n["c"].string_value == "9 15"

    assert n["d"]["e"].value == 9.7


def test_complex():
    t = complex_tree
    assert len(t["child"]) == 3
    assert len(t["child", 3]) == 2
    assert len(t["child", all_(lt(3))]) == 1
    assert len(t["child", any_(1)]) == 3
    assert len(t["child", any_(9)]) == 1
    assert len(t["child", any_(2)]) == 2
    assert len(t["dog"]["puppy"]) == 2
    assert len(t[startswith("chi") & endswith("ld")]) == 3
    assert len(t[startswith("chi") | startswith("do")]) == 4



from insights.core import CommandParser
from insights.tests import context_wrap
from insights.core.plugins import ContentException
import pytest

CMF = "blah: Command not found"
NO_SUCH_FILE = "/usr/bin/blah: No such file or directory"
MULTI_LINE = "blah: Command not found\n" \
             "/usr/bin/blah: No such file or directory"


class MockParser(CommandParser):
    def parse_content(self, content):
        self.data = content


def test_command_not_found():
    with pytest.raises(ContentException) as e:
        MockParser(context_wrap(CMF))
    assert "Command not found" in str(e.value)


def test_no_such_file_or_directory():
    with pytest.raises(ContentException) as e:
        MockParser(context_wrap(NO_SUCH_FILE))
    assert "No such file or directory" in str(e.value)


def test_multi_line():
    assert MULTI_LINE.split('\n') == MockParser(context_wrap(MULTI_LINE)).data



from insights.core.plugins import make_pass, make_fail
from insights.specs import Specs
from insights.plugins import rules_fixture_plugin
from insights.tests import InputData


UNAME = {
    'spec': Specs.uname,
    'data': """
Linux testbox.redhat.com 2.6.32-642.el6.x86_64 #1 SMP Tue Sep 16 01:56:35 EDT 2014 x86_64 x86_64 x86_64 GNU/Linux
""".strip()
}
RPMS = {
    'spec': Specs.installed_rpms,
    'path': '/etc/yum.repos.d/stuff',
    'data': """
kernel-2.6.32-573.el6.x86_64
bash-4.1.23-6.fc29.x86_64
rh-nginx112-nginx-1.12.1-2.el7.x86_64
""".strip()
}


def test_rules_fixture(run_rule):
    input_data = InputData('test_pass')
    input_data.add(UNAME['spec'], UNAME['data'])
    input_data.add(RPMS['spec'], RPMS['data'], path=RPMS['path'])
    expected = make_pass('PASS', bash_ver='bash-4.1.23-6.fc29', uname_ver='2.6.32')
    results = run_rule(rules_fixture_plugin.report, input_data)
    assert results == expected

    input_data = InputData('test_fail')
    input_data.add(RPMS['spec'], RPMS['data'], path=RPMS['path'])
    expected = make_fail('FAIL', bash_ver='bash-4.1.23-6.fc29', path=RPMS['path'])
    results = run_rule(rules_fixture_plugin.report, input_data)
    assert results == expected

    input_data = InputData('test_ret_none')
    results = run_rule(rules_fixture_plugin.report, input_data)
    assert results is None



import doctest
import pytest
from insights.parsers import lscpu, SkipException
from insights.tests import context_wrap


LSCPU_1 = """
Architecture:          x86_64
CPU op-mode(s):        32-bit, 64-bit
Byte Order:            Little Endian
CPU(s):                2
On-line CPU(s) list:   0,1
Thread(s) per core:    2
Core(s) per socket:    1
Socket(s):             1
NUMA node(s):          1
Vendor ID:             GenuineIntel
CPU family:            6
Model:                 60
Model name:            Intel Core Processor (Haswell, no TSX)
Stepping:              1
CPU MHz:               2793.530
BogoMIPS:              5587.06
Hypervisor vendor:     KVM
Virtualization type:   full
L1d cache:             32K
L1i cache:             32K
L2 cache:              4096K
NUMA node0 CPU(s):     0,1
""".strip()

LSCPU_2 = """
Architecture:          x86_64
CPU op-mode(s):        32-bit, 64-bit
Byte Order:            Little Endian
CPU(s):                2
On-line CPU(s) list:   0
Off-line CPU(s) list:  1
Thread(s) per core:    1
Core(s) per socket:    1
Socket(s):             1
NUMA node(s):          1
Vendor ID:             GenuineIntel
CPU family:            6
Model:                 60
Model name:            Intel Core Processor (Haswell, no TSX)
Stepping:              1
CPU MHz:               2793.530
BogoMIPS:              5587.06
Hypervisor vendor:     KVM
Virtualization type:   full
L1d cache:             32K
L1i cache:             32K
L2 cache:              4096K
NUMA node0 CPU(s):     0
""".strip()

BLANK = """
""".strip()

BAD_LSCPU = """
Architecture:          x86_64
CPU op-mode(s) =        32-bit, 64-bit
""".strip()


def test_lscpu_output():
    output = lscpu.LsCPU(context_wrap(LSCPU_1))
    assert output.info['Architecture'] == 'x86_64'
    assert len(output.info) == 22
    assert output.info['CPUs'] == '2'
    assert output.info['Threads per core'] == '2'
    assert output.info['Cores per socket'] == '1'
    assert output.info['Sockets'] == '1'

    output = lscpu.LsCPU(context_wrap(LSCPU_2))
    assert output.info['Architecture'] == 'x86_64'
    assert output.info['CPUs'] == '2'
    assert output.info['On-line CPUs list'] == '0'
    assert output.info['Off-line CPUs list'] == '1'
    assert output.info['Cores per socket'] == '1'
    assert output.info['Sockets'] == '1'


def test_lscpu_blank_output():
    with pytest.raises(SkipException) as e:
        lscpu.LsCPU(context_wrap(BLANK))
    assert "No data." in str(e)


def test_documentation():
    failed_count, tests = doctest.testmod(
        lscpu,
        globs={'output': lscpu.LsCPU(context_wrap(LSCPU_1))}
    )
    assert failed_count == 0



"""
InstalledRpms - Command ``rpm -qa``
===================================

The ``InstalledRpms`` class parses the output of the ``rpm -qa`` command.
Each line is parsed and stored in an ``InstalledRpm`` object.  The ``rpm -qa``
command may output data in different formats and each format can be
handled by the parsing routines of this class. The basic format of command is
the package and is shown in the Examples.

Sample input data::

    a52dec-0.7.4-18.el7.nux.x86_64  Tue 14 Jul 2015 09:25:38 AEST   1398536494
    aalib-libs-1.4.0-0.22.rc5.el7.x86_64    Tue 14 Jul 2015 09:25:40 AEST   1390535634
    abrt-2.1.11-35.el7.x86_64       Wed 09 Nov 2016 14:52:01 AEDT   1446193355
    ...
    kernel-3.10.0-230.el7synaptics.1186112.1186106.2.x86_64 Wed 20 May 2015 11:24:00 AEST   1425955944
    kernel-3.10.0-267.el7.x86_64    Sat 24 Oct 2015 09:56:17 AEDT   1434466402
    kernel-3.10.0-327.36.3.el7.x86_64       Wed 09 Nov 2016 14:53:25 AEDT   1476954923
    kernel-headers-3.10.0-327.36.3.el7.x86_64       Wed 09 Nov 2016 14:20:59 AEDT   1476954923
    kernel-tools-3.10.0-327.36.3.el7.x86_64 Wed 09 Nov 2016 15:09:42 AEDT   1476954923
    kernel-tools-libs-3.10.0-327.36.3.el7.x86_64    Wed 09 Nov 2016 14:52:13 AEDT   1476954923
    kexec-tools-2.0.7-38.el7_2.1.x86_64     Wed 09 Nov 2016 14:48:21 AEDT   1452845178
    ...
    zlib-1.2.7-15.el7.x86_64        Wed 09 Nov 2016 14:21:19 AEDT   1431443476
    zsh-5.0.2-14.el7_2.2.x86_64     Wed 09 Nov 2016 15:13:19 AEDT   1464185248


Examples:
    >>> type(rpms)
    <class 'insights.parsers.installed_rpms.InstalledRpms'>
    >>> 'openjpeg-libs' in rpms
    True
    >>> rpms.corrupt
    False
    >>> rpms.get_max('openjpeg-libs')
    0:openjpeg-libs-1.3-9.el6_3
    >>> type(rpms.get_max('openjpeg-libs'))
    <class 'insights.parsers.installed_rpms.InstalledRpm'>
    >>> rpms.get_min('openjpeg-libs')
    0:openjpeg-libs-1.3-9.el6_3
    >>> rpm = rpms.get_max('openssh-server')
    >>> rpm
    0:openssh-server-5.3p1-104.el6
    >>> type(rpm)
    <class 'insights.parsers.installed_rpms.InstalledRpm'>
    >>> rpm.package
    'openssh-server-5.3p1-104.el6'
    >>> rpm.nvr
    'openssh-server-5.3p1-104.el6'
    >>> rpm.source
    >>> rpm.name
    'openssh-server'
    >>> rpm.version
    '5.3p1'
    >>> rpm.release
    '104.el6'
    >>> rpm.arch
    'x86_64'
    >>> rpm.epoch
    '0'
    >>> from insights.parsers.installed_rpms import InstalledRpm
    >>> rpm2 = InstalledRpm.from_package('openssh-server-6.0-100.el6.x86_64')
    >>> rpm == rpm2
    False
    >>> rpm > rpm2
    False
    >>> rpm < rpm2
    True
"""
import json
import re
from collections import defaultdict

import six
import warnings

from ..util import rsplit
from .. import parser, get_active_lines, CommandParser
from .rpm_vercmp import rpm_version_compare
from insights.specs import Specs

# This list of architectures is taken from PDC (Product Definition Center):
# https://pdc.fedoraproject.org/rest_api/v1/arches/
KNOWN_ARCHITECTURES = [
    # Common architectures
    'x86_64',
    'i386',
    'i486',
    'i586',
    'i686',
    'src',
    'ia64',
    'ppc',
    'ppc64',
    's390',
    's390x',
    'amd64',
    '(none)',
    'noarch',
    # Less common
    'alpha',
    'alphaev4',
    'alphaev45',
    'alphaev5',
    'alphaev56',
    'alphaev6',
    'alphaev67',
    'alphaev68',
    'alphaev7',
    'alphapca56',
    'arm64',
    'armv5tejl',
    'armv5tel',
    'armv6l',
    'armv7hl',
    'armv7hnl',
    'armv7l',
    'athlon',
    'armhfp',
    'geode',
    'ia32e',
    'nosrc',
    'ppc64iseries',
    'ppc64le',
    'ppc64p7',
    'ppc64pseries',
    'sh3',
    'sh4',
    'sh4a',
    'sparc',
    'sparc64',
    'sparc64v',
    'sparcv8',
    'sparcv9',
    'sparcv9v',
    'aarch64',
]
"""list: List of recognized architectures.

This list is taken from the PDC (Product Definition Center) available
here https://pdc.fedoraproject.org/rest_api/v1/arches/.
"""


class RpmList(object):
    """
    Mixin class providing ``__contains__``, ``get_max``, ``get_min``,
    ``newest``, and ``oldest`` implementations for components that handle rpms.
    """

    def __contains__(self, package_name):
        """
        Checks if package name is in list of installed RPMs.

        Args:
            package_name (str): RPM package name such as 'bash'

        Returns:
            bool: True if package name is in list of installed packages, otherwise False
        """
        return package_name in self.packages

    def get_max(self, package_name):
        """
        Returns the highest version of the installed package with the given name.

        Args:
            package_name (str): Installed RPM package name such as 'bash'

        Returns:
            InstalledRpm: Installed RPM with highest version
        """
        if package_name not in self.packages:
            return None
        else:
            return max(self.packages[package_name])

    def get_min(self, package_name):
        """
        Returns the lowest version of the installed package with the given name.

        Args:
            package_name (str): Installed RPM package name such as 'bash'.

        Returns:
            InstalledRpm: Installed RPM with lowest version
        """
        if package_name not in self.packages:
            return None
        else:
            return min(self.packages[package_name])

    @property
    def is_hypervisor(self):
        """
        .. warning::
           This method is deprecated, please use
           :py:class:`insights.parsers.virt_what.VirtWhat` which uses the command `virt-what` to check the hypervisor type.

        bool: True if ".el[6|7]ev" exists in "vdsm".release, else False.
        """
        warnings.warn("`is_hypervisor` is deprecated: Use `virt_what.VirtWhat` which uses the command `virt-what` to check the hypervisor type.", DeprecationWarning)
        rpm = self.get_max("vdsm")
        return (True if rpm and rpm.release.endswith((".el6ev", ".el7ev")) else
                False)

    # re-export get_max/min with more descriptive names
    newest = get_max
    oldest = get_min


@parser(Specs.installed_rpms)
class InstalledRpms(CommandParser, RpmList):
    """
    A parser for working with data containing a list of installed RPM files on the system and
    related information.
    """
    def __init__(self, *args, **kwargs):
        self.errors = []
        """list: List of input lines that indicate an error acquiring the data on the client."""
        self.unparsed = []
        """list: List of input lines that raised an exception during parsing."""
        self.packages = defaultdict(list)
        """dict (InstalledRpm): Dictionary of RPMs keyed by package name."""

        super(InstalledRpms, self).__init__(*args, **kwargs)

    def parse_content(self, content):
        for line in get_active_lines(content, comment_char='COMMAND>'):
            if line.startswith('error:') or line.startswith('warning:'):
                self.errors.append(line)
            else:
                try:
                    # Try to parse from JSON input
                    rpm = InstalledRpm.from_json(line)
                    self.packages[rpm.name].append(rpm)
                except Exception:
                    # If that fails, try to parse from line input
                    if line.strip():
                        try:
                            rpm = InstalledRpm.from_line(line)
                            self.packages[rpm.name].append(rpm)
                        except Exception:
                            # Both ways failed
                            self.unparsed.append(line)
        # Don't want defaultdict's behavior after parsing is complete
        self.packages = dict(self.packages)

    @property
    def corrupt(self):
        """bool: True if RPM database is corrupted, else False."""
        return any('rpmdbNextIterator' in s for s in self.errors)


p = re.compile(r"(\d+|[a-z]+|\.|-|_)")


def _int_or_str(c):
    try:
        return int(c)
    except ValueError:
        return c


def vcmp(s):
    return [_int_or_str(c) for c in p.split(s) if c and c not in (".", "_", "-")]


def pad_version(left, right):
    """Returns two sequences of the same length so that they can be compared.
    The shorter of the two arguments is lengthened by inserting extra zeros
    before non-integer components.  The algorithm attempts to align character
    components."""
    pair = vcmp(left), vcmp(right)

    mn, mx = min(pair, key=len), max(pair, key=len)

    for idx, c in enumerate(mx):

        try:
            a = mx[idx]
            b = mn[idx]
            if type(a) != type(b):
                mn.insert(idx, 0)
        except IndexError:
            if type(c) is int:
                mn.append(0)
            elif isinstance(c, six.string_types):
                mn.append('')
            else:
                raise Exception("pad_version failed (%s) (%s)" % (left, right))

    return pair


class InstalledRpm(object):
    """
    Class for holding information about one installed RPM.

    This class is usually created from dictionary with following structure::

         {
            'name': 'package name',
            'version': 'package version',
            'release': 'package release',
            'arch': 'package architecture'
          }

    It may also contain supplementary information from SOS report or epoch
    information from JSON.

    Factory methods are provided such as ``from_package`` to create an object from a
    short package string::

        kernel-devel-3.10.0-327.36.1.el7.x86_64

    ``from_json`` to create an object from JSON::

       {"name": "kernel-devel",
        "version": "3.10.0",
        "release": "327.36.1.el7",
        "arch": "x86_64"}

    and ``from_line`` to create an object from a long package string:

        .. code:: python

            ('kernel-devel-3.10.0-327.36.1.el7.x86_64'
             '                                '
             'Wed May 18 14:16:21 2016' '\t'
             '1410968065' '\t'
             'Red Hat, Inc.' '\t'
             'hs20-bc2-4.build.redhat.com' '\t'
             '8902150305004...b3576ff37da7e12e2285358267495ac48a437d4eefb3213' '\t'
             'RSA/8, Mon Aug 16 11:14:17 2010, Key ID 199e2f91fd431d51')
    """
    PRODUCT_SIGNING_KEYS = [
        'F76F66C3D4082792', '199e2f91fd431d51', '5326810137017186',
        '45689c882fa658e0', '219180cddb42a60e', '7514f77d8366b0d9',
        'fd372689897da07a', '938a80caf21541eb',
        '08b871e6a5787476',
        'E191DDB2C509E861'
    ]
    """
    list: List of package-signing keys. Should be updated timely according to
          https://access.redhat.com/security/team/key/
    """
    SOSREPORT_KEYS = [
        'installtime', 'buildtime', 'vendor', 'buildserver', 'pgpsig', 'pgpsig_short'
    ]
    """list: List of keys for SOS Report RPM information."""

    def __init__(self, data):
        self.name = None
        """str: RPM package name."""
        self.version = None
        """str: RPM package version."""
        self.release = None
        """str: RPM package release."""
        self.arch = None
        """str: RPM package architecture."""
        self.redhat_signed = None
        """bool: True when RPM package is signed by Red Hat, False when RPM package is not signed by Red Hat,
        None when no sufficient info to determine"""

        if isinstance(data, six.string_types):
            data = self._parse_package(data)

        for k, v in data.items():
            setattr(self, k, v)
        self.epoch = data['epoch'] if 'epoch' in data and data['epoch'] != '(none)' else '0'
        _gpg_key_pos = data.get('sigpgp', data.get('rsaheader', data.get('pgpsig_short', data.get('pgpsig', data.get('vendor', '')))))
        if _gpg_key_pos:
            self.redhat_signed = any(key in _gpg_key_pos for key in self.PRODUCT_SIGNING_KEYS)

    @classmethod
    def from_package(cls, package_string):
        """
        The object of this class is usually created from dictionary. Alternatively it can be
        created from package string.

        Args:
            package_string (str): package string in the following format (shown as Python string)::

                'kernel-devel-3.10.0-327.36.1.el7.x86_64'
        """
        return cls(cls._parse_package(package_string))

    @classmethod
    def from_json(cls, json_line):
        """
        The object of this class is usually created from dictionary. Alternatively it can be
        created from JSON line.

        Args:
            json_line (str): JSON string in the following format (shown as Python string)::

                '{"name": "kernel-devel", "version": "3.10.0", "release": "327.36.1.el7", "arch": "x86_64"}'
        """
        return cls(json.loads(json_line))

    @classmethod
    def from_line(cls, line):
        """
        The object of this class is usually created from dictionary. Alternatively it can be
        created from package line.

        Args:
            line (str): package line in the following format (shown as Python string):

                .. code-block:: python

                    ('kernel-devel-3.10.0-327.36.1.el7.x86_64'
                     '                                '
                     'Wed May 18 14:16:21 2016' '\t'
                     '1410968065' '\t'
                     'Red Hat, Inc.' '\t'
                     'hs20-bc2-4.build.redhat.com' '\t'
                     '8902150305004...b3576ff37da7e12e2285358267495ac48a437d4eefb3213' '\t'
                     'RSA/8, Mon Aug 16 11:14:17 2010, Key ID 199e2f91fd431d51')

        """
        return cls(cls._parse_line(line))

    @staticmethod
    def _arch_sep(package_string):
        """
        Helper method for finding if arch separator is '.' or '-'

        Args:
            package_string (str): dash separated package string such as 'bash-4.2.39-3.el7'.

        Returns:
            str: arch separator
        """
        return '.' if package_string.rfind('.') > package_string.rfind('-') else '-'

    @classmethod
    def _parse_package(cls, package_string):
        """
        Helper method for parsing package string.

        Args:
            package_string (str): dash separated package string such as 'bash-4.2.39-3.el7'

        Returns:
            dict: dictionary containing 'name', 'version', 'release' and 'arch' keys
        """
        pkg, arch = rsplit(package_string, cls._arch_sep(package_string))
        if arch not in KNOWN_ARCHITECTURES:
            pkg, arch = (package_string, None)
        pkg, release = rsplit(pkg, '-')
        name, version = rsplit(pkg, '-')
        epoch, version = version.split(':', 1) if ":" in version else ['0', version]
        # oracleasm packages have a dash in their version string, fix that
        if name.startswith('oracleasm') and name.endswith('.el5'):
            name, version2 = name.split('-', 1)
            version = version2 + '-' + version
        return {
            'name': name,
            'version': version,
            'release': release,
            'arch': arch,
            'epoch': epoch
        }

    @classmethod
    def _parse_line(cls, line):
        """
        Helper method for parsing package line with or without SOS report information.

        Args:
            line (str): package line with or without SOS report information

        Returns:
            dict: dictionary containing 'name', 'version', 'release' and 'arch' keys plus
                  additionally 'installtime', 'buildtime', 'vendor', 'buildserver', 'pgpsig',
                  'pgpsig_short' if these are present.
        """
        try:
            pkg, rest = line.split(None, 1)
        except ValueError:
            rpm = cls._parse_package(line.strip())
            return rpm
        rpm = cls._parse_package(pkg)
        rest = rest.split('\t')
        for i, value in enumerate(rest):
            rpm[cls.SOSREPORT_KEYS[i]] = value
        return rpm

    @property
    def package(self):
        """str: Package `name-version-release` string."""
        return u'{0}-{1}-{2}'.format(self.name,
                                     self.version,
                                     self.release)

    @property
    def package_with_epoch(self):
        """
        str: Package string in the format::

            name-epoch:version-release
        """
        return u'{0}-{1}:{2}-{3}'.format(self.name,
                                         self.epoch,
                                         self.version,
                                         self.release)

    @property
    def nvr(self):
        """str: Package `name-version-release` string."""
        return self.package

    @property
    def nvra(self):
        """str: Package `name-version-release.arch` string."""
        return ".".join([self.package, self.arch])

    @property
    def nevra(self):
        """
        str: Package string in the format::

            name-epoch:version-release.arch
        """
        return ".".join([self.package_with_epoch, self.arch])

    @property
    def source(self):
        """InstalledRpm: Returns source RPM of this RPM object."""
        if hasattr(self, 'srpm'):
            rpm = self.from_package(self.srpm)
            # Source RPMs don't have epoch for some reason
            rpm.epoch = self.epoch
            return rpm

    def __getitem__(self, item):
        """
        Allows to use `rpm["element"]` instead of `rpm.element`. Dot notation should be preferred,
        however it is especially useful for values containing dash, such as "pgpsig_short".
        """
        return getattr(self, item)

    def __str__(self):
        return '{0}:{1}'.format(self.epoch, self.package)

    def __unicode__(self):
        return str(self)

    def __repr__(self):
        return str(self)

    def __eq__(self, other):
        if not isinstance(other, InstalledRpm):
            return False

        if self.name != other.name:
            raise ValueError('Cannot compare packages with differing names {0} != {1}'
                             .format(self.name, other.name))

        return rpm_version_compare(self, other) == 0

    def __lt__(self, other):
        if not isinstance(other, InstalledRpm):
            return False

        if self == other:
            return False

        return rpm_version_compare(self, other) < 0

    def __ne__(self, other):
        return not self == other

    def __gt__(self, other):
        return isinstance(other, InstalledRpm) and other.__lt__(self)

    def __ge__(self, other):
        return isinstance(other, InstalledRpm) and not self.__lt__(other)

    def __le__(self, other):
        return isinstance(other, InstalledRpm) and not other.__lt__(self)

    def __hash__(self):
        # Python 3 requires hash implementation to have hashable object.
        try:
            # Just NVR is not enouch for uniqueness. Try NVRA first.
            value = self.nvra
        except TypeError:
            value = self.nvr
        return hash(value)


# re-exports
from_package = InstalledRpm.from_package
Rpm = InstalledRpm
Installed = InstalledRpms



from collections import defaultdict
from insights.core import filters

from insights.specs import Specs
from insights.specs.default import DefaultSpecs


def setup_function(func):
    if func is test_get_filter:
        filters.add_filter(Specs.ps_aux, "COMMAND")

    if func is test_get_filter_registry_point:
        filters.add_filter(Specs.ps_aux, "COMMAND")
        filters.add_filter(DefaultSpecs.ps_aux, "MEM")

    if func is test_filter_dumps_loads:
        filters.add_filter(Specs.ps_aux, "COMMAND")


def teardown_function(func):
    if func is test_get_filter:
        del filters.FILTERS[Specs.ps_aux]

    if func is test_get_filter_registry_point:
        del filters.FILTERS[Specs.ps_aux]
        del filters.FILTERS[DefaultSpecs.ps_aux]

    if func is test_filter_dumps_loads:
        del filters.FILTERS[Specs.ps_aux]


def test_filter_dumps_loads():
    r = filters.dumps()
    assert r is not None

    filters.FILTERS = defaultdict(set)
    filters.loads(r)

    assert Specs.ps_aux in filters.FILTERS
    assert filters.FILTERS[Specs.ps_aux] == set(["COMMAND"])


def test_get_filter():
    f = filters.get_filters(Specs.ps_aux)
    assert "COMMAND" in f

    f = filters.get_filters(DefaultSpecs.ps_aux)
    assert "COMMAND" in f


def test_get_filter_registry_point():
    s = set(["COMMAND", "MEM"])
    f = filters.get_filters(DefaultSpecs.ps_aux)
    assert f & s == s

    f = filters.get_filters(Specs.ps_aux)
    assert "COMMAND" in f
    assert "MEM" not in f



from __future__ import print_function
import copy
import inspect
import itertools
import json
import logging
import six
import six.moves
from collections import defaultdict
from functools import wraps
from operator import eq

try:
    from StringIO import StringIO
except ImportError:
    from io import StringIO

import insights
from insights import apply_filters
from insights.core import dr, filters, spec_factory
from insights.core.context import Context
from insights.core.spec_factory import RegistryPoint
from insights.specs import Specs


# we intercept the add_filter call during integration testing so we can ensure
# that rules add filters to datasources that *should* be filterable
ADDED_FILTERS = defaultdict(set)
add_filter = filters.add_filter
find = spec_factory.find


def _intercept_add_filter(func):
    @wraps(func)
    def inner(ds, pattern):
        ret = add_filter(ds, pattern)
        calling_module = inspect.stack()[1][0].f_globals.get("__name__")
        ADDED_FILTERS[calling_module].add(ds)
        return ret
    return inner


def _intercept_find(func):
    @wraps(func)
    def inner(ds, pattern):
        ret = find(ds, pattern)
        calling_module = inspect.stack()[1][0].f_globals.get("__name__")
        ADDED_FILTERS[calling_module].add(ds)
        return ret
    return inner


filters.add_filter = _intercept_add_filter(filters.add_filter)
insights.add_filter = _intercept_add_filter(insights.add_filter)

spec_factory.find = _intercept_find(spec_factory.find)


def _get_registry_points(component):
    results = set()
    for c in dr.walk_tree(component):
        try:
            if isinstance(c, RegistryPoint):
                results.add(c)
        except:
            pass
    return results


logger = logging.getLogger(__name__)

ARCHIVE_GENERATORS = []
HEARTBEAT_ID = "99e26bb4823d770cc3c11437fe075d4d1a4db4c7500dad5707faed3b"
HEARTBEAT_NAME = "insights-heartbeat-9cd6f607-6b28-44ef-8481-62b0e7773614"

DEFAULT_RELEASE = "Red Hat Enterprise Linux Server release 7.2 (Maipo)"
DEFAULT_HOSTNAME = "hostname.example.com"


def deep_compare(result, expected):
    """
    Deep compare rule reducer results when testing.
    """
    logger.debug("--Comparing-- (%s) %s to (%s) %s", type(result), result, type(expected), expected)

    if isinstance(result, dict) and expected is None:
        assert result["type"] == "skip", result
        return

    assert eq(result, expected)


def run_input_data(component, input_data):
    broker = dr.Broker()
    for k, v in input_data.data.items():
        broker[k] = v

    graph = dr.get_dependency_graph(component)
    broker = dr.run(graph, broker=broker)
    for v in broker.tracebacks.values():
        print(v)
    return broker


def run_test(component, input_data, expected=None):
    if filters.ENABLED:
        mod = component.__module__
        sup_mod = '.'.join(mod.split('.')[:-1])
        rps = _get_registry_points(component)
        filterable = set(d for d in rps if dr.get_delegate(d).filterable)
        missing_filters = filterable - ADDED_FILTERS.get(mod, set()) - ADDED_FILTERS.get(sup_mod, set())
        if missing_filters:
            names = [dr.get_name(m) for m in missing_filters]
            msg = "%s must add filters to %s"
            raise Exception(msg % (mod, ", ".join(names)))

    broker = run_input_data(component, input_data)
    if expected:
        deep_compare(broker.get(component), expected)
    return broker.get(component)


def integrate(input_data, component):
    return run_test(component, input_data)


def context_wrap(lines,
                 path="path",
                 hostname=DEFAULT_HOSTNAME,
                 release=DEFAULT_RELEASE,
                 version="-1.-1",
                 machine_id="machine_id",
                 strip=True,
                 split=True,
                 **kwargs):
    if isinstance(lines, six.string_types):
        if strip:
            lines = lines.strip()
        if split:
            lines = lines.splitlines()

    return Context(content=lines,
                   path=path, hostname=hostname,
                   release=release, version=version.split("."),
                   machine_id=machine_id, relative_path=path, **kwargs)


input_data_cache = {}


counter = itertools.count()


def create_metadata(system_id, product):
    ctx_metadata = {
        "system_id": system_id,
        "links": []
    }
    ctx_metadata["type"] = product.role
    ctx_metadata["product"] = product.__class__.__name__
    return json.dumps(ctx_metadata)


class InputData(object):
    """
    Helper class used with integrate. The role of this class is to represent
    data files sent to parsers and rules in a format similar to what lays on
    disk.

    Example Usage::

        input_data = InputData()
        input_data.add("messages", "this is some messages content")
        input_data.add("uname", "this is some uname content")

    If `release` is specified when InputData is constructed, it will be
    added to every *mock* record when added. This is useful for testing parsers
    that rely on context.release.

    If `path` is specified when calling the `add` method, the record will
    contain the specified value in the context.path field.  This is useful for
    testing pattern-like file parsers.
    """
    def __init__(self, name=None, hostname=None):
        cnt = input_data_cache.get(name, 0)
        self.name = "{0}-{1:0>5}".format(name, cnt)
        self.data = {}
        input_data_cache[name] = cnt + 1
        if hostname:
            self.add(Specs.hostname, hostname)

    def __setitem__(self, key, value):
        self.add(key, value)

    def __getitem__(self, key):
        return self.data[key]

    def _make_path(self):
        return str(next(counter)) + "BOGUS"

    def get(self, key, default):
        return self.data.get(key, default)

    def items(self):
        return self.data.items()

    def clone(self, name):
        the_clone = InputData(name)
        the_clone.data = {}
        for k, v in self.data.items():
            the_clone.data[k] = copy.deepcopy(v)
        return the_clone

    def add_component(self, comp, obj):
        """
        Allow adding arbitrary objects as components. This allows tests to mock
        components that have external dependencies so their dependents can be
        integration tested.
        """
        self.data[comp] = obj

    def add(self, spec, content, path=None, do_filter=True):
        if not path:  # path must change to allow parsers to fire
            path = self._make_path()
        if not path.startswith("/"):
            path = "/" + path

        if dr.get_delegate(spec).raw:
            content_iter = content
        else:
            if not isinstance(content, list):
                content_iter = [l.rstrip() for l in StringIO(content).readlines()]
            else:
                content_iter = content

            if do_filter:
                content_iter = list(apply_filters(spec, content_iter))

        content_provider = context_wrap(content_iter, path=path, split=False)
        if dr.get_delegate(spec).multi_output:
            if spec not in self.data:
                self.data[spec] = []
            self.data[spec].append(content_provider)
        else:
            self.data[spec] = content_provider

        return self

    def __repr__(self):
        if self.name:
            return "<InputData {name:%s}>" % (self.name)
        else:
            return super(InputData, self).__repr__()


# Helper constants when its necessary to test for a specific RHEL major version
# eg RHEL6, but the minor version isn't important
RHEL4 = "Red Hat Enterprise Linux AS release 4 (Nahant Update 9)"
RHEL5 = "Red Hat Enterprise Linux Server release 5.11 (Tikanga)"
RHEL6 = "Red Hat Enterprise Linux Server release 6.5 (Santiago)"
RHEL7 = "Red Hat Enterprise Linux Server release 7.0 (Maipo)"
RHEL8 = "Red Hat Enterprise Linux release 8.0 (Ootpa)"


def redhat_release(major, minor=""):
    """
    Helper function to construct a redhat-release string for a specific RHEL
    major and minor version.  Only constructs redhat-releases for RHEL major
    releases 4, 5, 6 & 7

    Arguments:
        major: RHEL major number. Accepts str, int or float (as major.minor)
        minor: RHEL minor number. Optional and accepts str or int

    For example, to construct a redhat-release for::

        RHEL4U9:  redhat_release('4.9') or (4.9) or (4, 9)
        RHEL5 GA: redhat_release('5')   or (5.0) or (5, 0) or (5)
        RHEL6.6:  redhat_release('6.6') or (6.6) or (6, 6)
        RHEL7.1:  redhat_release('7.1') or (7.1) or (7, 1)

    Limitation with float args: (x.10) will be parsed as minor = 1
    """
    if isinstance(major, str) and "." in major:
        major, minor = major.split(".")
    elif isinstance(major, float):
        major, minor = str(major).split(".")
    elif isinstance(major, int):
        major = str(major)
    if isinstance(minor, int):
        minor = str(minor)

    if major == "4":
        if minor:
            minor = "" if minor == "0" else " Update %s" % minor
        return "Red Hat Enterprise Linux AS release %s (Nahant%s)" % (major, minor)

    if major == "8":
        if not minor:
            minor = "0"
        return "Red Hat Enterprise Linux release %s.%s Beta (Ootpa)" % (major, minor)

    template = "Red Hat Enterprise Linux Server release %s%s (%s)"
    if major == "5":
        if minor:
            minor = "" if minor == "0" else "." + minor
        return template % (major, minor, "Tikanga")
    elif major == "6" or major == "7":
        if not minor:
            minor = "0"
        name = "Santiago" if major == "6" else "Maipo"
        return template % (major, "." + minor, name)
    else:
        raise Exception("invalid major version: %s" % major)


def archive_provider(component, test_func=deep_compare, stride=1):
    """
    Decorator used to register generator functions that yield InputData and
    expected response tuples.  These generators will be consumed by py.test
    such that:

    - Each InputData will be passed into an integrate() function
    - The result will be compared [1] against the expected value from the
      tuple.

    Parameters
    ----------
    component: (str)
        The component to be tested.
    test_func: function
        A custom comparison function with the parameters (result, expected).
        This will override the use of the compare() [1] function.
    stride: int
        yield every `stride` InputData object rather than the full set. This
        is used to provide a faster execution path in some test setups.

    [1] insights.tests.deep_compare()
    """
    def _wrap(func):
        @six.wraps(func)
        def __wrap(stride=stride):
            for input_data, expected in itertools.islice(func(), None, None, stride):
                yield component, test_func, input_data, expected

        __wrap.stride = stride
        ARCHIVE_GENERATORS.append(__wrap)
        return __wrap
    return _wrap



"""
saphostexec - Commands
======================

Shared parsers for parsing output of the ``saphostexec [option]`` commands.

SAPHostExecStatus- command ``saphostexec -status``
--------------------------------------------------

SAPHostExecVersion - command ``saphostexec -version``
-----------------------------------------------------
"""
from .. import parser, CommandParser, LegacyItemAccess
from insights.parsers import SkipException
from insights.specs import Specs
from collections import namedtuple


@parser(Specs.saphostexec_status)
class SAPHostExecStatus(CommandParser, LegacyItemAccess):
    """
    Class for parsing the output of `saphostexec -status` command.

    Typical output of the command is::

        saphostexec running (pid = 9159)
        sapstartsrv running (pid = 9163)
        saposcol running (pid = 9323)

    Attributes:
        is_running (bool): The SAP Host Agent is running or not.
        services (list): List of services.

    Examples:
        >>> type(sha_status)
        <class 'insights.parsers.saphostexec.SAPHostExecStatus'>
        >>> sha_status.is_running
        True
        >>> sha_status.services['saphostexec']
        '9159'
    """

    def parse_content(self, content):
        self.is_running = False
        self.services = self.data = {}
        if 'saphostexec stopped' not in content[0]:
            for line in content:
                line_splits = line.split()
                self.services[line_splits[0]] = ''
                if len(line_splits) == 5 and line_splits[1] == 'running':
                    self.services[line_splits[0]] = line_splits[-1][:-1]
                else:
                    raise SkipException("Incorrect status: '{0}'".format(line))

            self.is_running = self.services and all(p for p in self.services.values())


@parser(Specs.saphostexec_version)
class SAPHostExecVersion(CommandParser, LegacyItemAccess):
    """
    Class for parsing the output of `saphostexec -version` command.

    Typical output of the command is::

        *************************** Component ********************
        /usr/sap/hostctrl/exe/saphostexec: 721, patch 1011, changelist 1814854, linuxx86_64, opt (Jan 13 2018, 04:43:56)
        /usr/sap/hostctrl/exe/sapstartsrv: 721, patch 1011, changelist 1814854, linuxx86_64, opt (Jan 13 2018, 04:43:56)
        /usr/sap/hostctrl/exe/saphostctrl: 721, patch 1011, changelist 1814854, linuxx86_64, opt (Jan 13 2018, 04:43:56)
        /usr/sap/hostctrl/exe/xml71d.so: 721, patch 1011, changelist 1814854, linuxx86_64, opt (Jan 13 2018, 01:12:10)
        **********************************************************
        --------------------
        SAPHOSTAGENT information
        --------------------
        kernel release                721
        kernel make variant           721_REL
        compiled on                   Linux GNU SLES-9 x86_64 cc4.1.2  for linuxx86_64
        compiled for                  64 BIT
        compilation mode              Non-Unicode
        compile time                  Jan 13 2018 04:40:52
        patch number                  33
        latest change number          1814854
        ---------------------
        supported environment
        ---------------------
        operating system
        Linux 2.6
        Linux 3
        Linux


    Attributes:
        components (dict): Dict of :py:class:`SAPComponent` instances.

    Examples:
        >>> type(sha_version)
        <class 'insights.parsers.saphostexec.SAPHostExecVersion'>
        >>> sha_version.components['saphostexec'].version
        '721'
        >>> sha_version.components['saphostexec'].patch
        '1011'
    """

    SAPComponent = namedtuple("SAPComponent",
            field_names=["version", "patch", "changelist"])
    """namedtuple: Type for storing the SAP components"""

    def parse_content(self, content):
        self.components = self.data = {}
        for line in content:
            # Only process component lines for now
            if not line.startswith('/usr/sap/hostctrl/exe/'):
                continue
            key, val = line.split(':', 1)
            key = key.split('/')[-1]
            ver, pch, chl, _ = [s.split()[-1].strip() for s in val.split(', ', 3)]
            self.components[key] = self.SAPComponent(ver, pch, chl)



"""
Units Manged By Systemctl (services)
====================================

Parsers included in this module are:

ListUnits - command ``/bin/systemctl list-units``
-------------------------------------------------

UnitFiles - command ``/bin/systemctl list-unit-files``
------------------------------------------------------

"""
from .. import get_active_lines
from ... import Parser, parser
from insights.specs import Specs


@parser(Specs.systemctl_list_unit_files)
class UnitFiles(Parser):
    """
    The UnitFiles class parses the output of ``/bin/systemctl list-unit-files`` and provides
    information about enabled services.

    Output of Command::

        mariadb.service                               enabled
        neutron-openvswitch-agent.service             enabled
        neutron-ovs-cleanup.service                   enabled
        neutron-server.service                        enabled
        runlevel0.target                              disabled
        runlevel1.target                              disabled
        runlevel2.target                              enabled

    Example:

        >>> conf = shared[UnitFiles]
        >>> conf.is_on('existing-enabled-service.service')
        True
        >>> conf.is_on('existing-disabled-service.service')
        False
        >>> conf.is_on('nonexistent-service.service')
        False
        >>> conf.exists('existing-enabled-service.service')
        True
        >>> conf.exists('existing-disabled-service.service')
        True
        >>> conf.exists('nonexistent-service.service')
        False
        >>> 'existing-enabled-service.service' in conf.services
        True
        >>> 'existing-disabled-service.service' in conf.services
        True
        >>> 'nonexistent-service.service' in conf.services
        False
        >>> conf.services['existing-enabled-service.service']
        True
        >>> conf.services['existing-disabled-service.service']
        False
        >>> conf.services['nonexistent-service.service']
        KeyError: 'nonexistent-service.service'
    """

    def __init__(self, *args, **kwargs):
        self.services = {}
        """dict: Dictionary of bool indicating if service is enabled,
        access by service name ."""
        self.service_list = []
        """list: List of service names in order of appearance."""
        self.parsed_lines = {}
        """dict: Dictionary of content lines access by service name."""
        super(UnitFiles, self).__init__(*args, **kwargs)

    def parse_content(self, content):
        """
        Main parsing class method which stores all interesting data from the content.

        Args:
            content (context.content): Parser context content
        """
        # 'static' means 'on' to fulfill dependency of something else that is on
        # man systemctl - "is-enabled" knows these states
        valid_states = set(['enabled', 'enabled-runtime', 'linked', 'linked-runtime', 'masked',
                            'masked-runtime', 'static', 'indirect', 'disabled', 'generated',
                            'transient', 'bad', 'invalid'])
        # man systemctl - "is-enabled" considers these to be enabled
        on_states = set(['enabled', 'enabled-runtime', 'static', 'indirect', 'generated', 'transient'])

        for line in get_active_lines(content):
            parts = line.split(None)  # AWK like split, strips whitespaces
            if len(parts) == 2 and any(part in valid_states for part in parts):
                service, state = parts
                enabled = state in on_states
                self.services[service] = enabled
                self.parsed_lines[service] = line
                self.service_list.append(service)

    def is_on(self, service_name):
        """
        Checks if the service is enabled in systemctl.

        Args:
            service_name (str): service name including '.service'

        Returns:
            Union[bool, None]: True if service is enabled, False if it is disabled. None if the
            service doesn't exist.
        """
        return self.services.get(service_name, None)

    def exists(self, service_name):
        """
        Checks if the service is listed in systemctl.

        Args:
           service_name (str): service name including '.service'

        Returns:
            bool: True if service exists, False otherwise.
        """
        return service_name in self.service_list


@parser(Specs.systemctl_list_units)
class ListUnits(Parser):

    """
    The ListUnits class parses the output of ``/bin/systemctl list-units`` and provides
    information about all the services listed under it.

    Output of Command::

        sockets.target                      loaded active active    Sockets
        swap.target                         loaded active active    Swap
        systemd-shutdownd.socket            loaded active listening Delayed Shutdown Socket
        neutron-dhcp-agent.service          loaded active running   OpenStack Neutron DHCP Agent
        neutron-openvswitch-agent.service   loaded active running   OpenStack Neutron Open vSwitch Agent

    Example:

        >>> units.get_service_details('swap.target')
        {'LOAD': 'loaded', 'ACTIVE': 'active', 'SUB': 'active', 'UNIT': 'swap.target'}
        >>> units.unit_list['swap.target']
        {'LOAD': 'loaded', 'ACTIVE': 'active', 'SUB': 'active', 'UNIT': 'swap.target'}
        >>> units.is_active('swap.target')
        True
        >>> units.unit_list['random.service']
        {'LOAD': None, 'ACTIVE': None, 'SUB': None, 'UNIT': None}
    """
    def __init__(self, *args, **kwargs):
        self.unit_list = {}
        """dict: Dictionary service detail like active, running, exited, dead"""
        super(ListUnits, self).__init__(*args, **kwargs)

    def parse_service_details(self, parts):
        # man systemctl - "is-enabled" knows these states
        valid_states = set(['invalid', 'loaded', 'inactive', 'active',
                            'exited', 'running', 'failed', 'mounted', 'waiting', 'plugged'])

        valid_units = set(['service', 'socket', 'device', 'mount', 'automount', 'swap', 'target',
                           'path', 'timer', 'slice', 'scope'])

        if any(part in valid_states for part in parts):
            service_details = {}
            service_details['LOAD'] = parts[1]
            service_details['ACTIVE'] = parts[2]
            service_details['SUB'] = parts[3]
            if any(unit in parts[0] for unit in valid_units):
                service_details['UNIT'] = parts[0]
                return service_details

    def parse_content(self, content):
        """
        Main parsing class method which stores all interesting data from the content.

        Args:
            content (context.content): Parser context content
        """
        BULLET_CHAR_U = u'\u25CF'
        BULLET_CHAR_B = b"\xe2\x97\x8f"
        for line in get_active_lines(content):
            parts = line.split(None)  # AWK like split, strips whitespaces
            if parts[0] == BULLET_CHAR_U or parts[0].encode('utf-8') == BULLET_CHAR_B or parts[0] == '*':
                self.unit_list[parts[1]] = self.parse_service_details(parts[1:])
            else:
                self.unit_list[parts[0]] = self.parse_service_details(parts)

    def get_service_details(self, service_name):
        """
        Return the service details collected by systemctl.

        Args:
            service_name (str): service name including its extension.

        Returns:
            dict: Dictionary containing details for the service.
            if service is not present dictonary values will be `None`::

            {'LOAD': 'loaded', 'ACTIVE': 'active', 'SUB': 'running', 'UNIT': 'neutron-dhcp-agent.service'}
        """
        empty_details = {'LOAD': None, 'ACTIVE': None, 'SUB': None, 'UNIT': None}
        return self.unit_list.get(service_name, empty_details)

    def is_loaded(self, service_name):
        """
        Return the LOAD state of service managed by systemd.

        Args:
            service_name (str): service name including its extension.

        Returns:
            bool: True if service is loaded False if not loaded
        """
        return self.get_service_details(service_name)['LOAD'] == 'loaded'

    def is_active(self, service_name):
        """
        Return the ACTIVE state of service managed by systemd.

        Args:
            service_name (str): service name including its extension.

        Returns:
            bool: True if service is active False if inactive
        """
        return self.get_service_details(service_name)['ACTIVE'] == 'active'

    def is_running(self, service_name):
        """
        Return the SUB state of service managed by systemd.

        Args:
            service_name (str): service name including its extension.

        Returns:
            bool: True if service is running False in all other states.
        """
        return self.get_service_details(service_name)['SUB'] == 'running'



import doctest
import pytest
from insights.parsers import ParseException
from insights.tests import context_wrap
from insights.parsers import db2licm
from insights.parsers.db2licm import DB2Info

INVALID_OUTPUT = "".strip()

VALID_OUTPUT = """
Product name:                     DB2 Enterprise Server Edition
License type:                     CPU Option
Expiry date:                      Permanent
Product identifier:               db2ese
Version information:              9.7
Enforcement policy:               Soft Stop
Features:
DB2 Performance Optimization ESE: Not licensed
DB2 Storage Optimization:         Not licensed
DB2 Advanced Access Control:      Not licensed
IBM Homogeneous Replication ESE:  Not licensed
""".strip()

VALID_OUTPUT_MULTIPLE = """
Product name:                     DB2 Enterprise Server Edition
License type:                     CPU Option
Expiry date:                      Permanent
Product identifier:               db2ese
Version information:              9.7
Enforcement policy:               Soft Stop
Features:
DB2 Performance Optimization ESE: Not licensed
DB2 Storage Optimization:         Not licensed
DB2 Advanced Access Control:      Not licensed
IBM Homogeneous Replication ESE:  Not licensed

Product name:                     DB2 Connect Server
Expiry date:                      Expired
Product identifier:               db2consv
Version information:              9.7
Concurrent connect user policy:   Disabled
Enforcement policy:               Soft Stop
""".strip()


def test_valid_command_output_1():
    parser_result = DB2Info(context_wrap(VALID_OUTPUT))
    assert parser_result is not None

    assert parser_result["DB2 Enterprise Server Edition"]["License type"] == "CPU Option"
    assert parser_result["DB2 Enterprise Server Edition"]["Expiry date"] == "Permanent"
    assert parser_result["DB2 Enterprise Server Edition"]["Product identifier"] == "db2ese"
    assert parser_result["DB2 Enterprise Server Edition"]["Version information"] == "9.7"
    assert parser_result["DB2 Enterprise Server Edition"]["Enforcement policy"] == "Soft Stop"
    assert parser_result["DB2 Enterprise Server Edition"]["DB2 Performance Optimization ESE"] == "Not licensed"
    assert parser_result["DB2 Enterprise Server Edition"]["DB2 Storage Optimization"] == "Not licensed"
    assert parser_result["DB2 Enterprise Server Edition"]["DB2 Advanced Access Control"] == "Not licensed"
    assert parser_result["DB2 Enterprise Server Edition"]["IBM Homogeneous Replication ESE"] == "Not licensed"


def test_valid_command_output_2():
    parser_result = DB2Info(context_wrap(VALID_OUTPUT_MULTIPLE))
    assert parser_result is not None

    assert parser_result["DB2 Enterprise Server Edition"]["License type"] == "CPU Option"
    assert parser_result["DB2 Enterprise Server Edition"]["Expiry date"] == "Permanent"
    assert parser_result["DB2 Enterprise Server Edition"]["Product identifier"] == "db2ese"
    assert parser_result["DB2 Enterprise Server Edition"]["Version information"] == "9.7"
    assert parser_result["DB2 Enterprise Server Edition"]["Enforcement policy"] == "Soft Stop"
    assert parser_result["DB2 Enterprise Server Edition"]["DB2 Performance Optimization ESE"] == "Not licensed"
    assert parser_result["DB2 Enterprise Server Edition"]["DB2 Storage Optimization"] == "Not licensed"
    assert parser_result["DB2 Enterprise Server Edition"]["DB2 Advanced Access Control"] == "Not licensed"
    assert parser_result["DB2 Enterprise Server Edition"]["IBM Homogeneous Replication ESE"] == "Not licensed"

    assert parser_result["DB2 Connect Server"]["Expiry date"] == "Expired"
    assert parser_result["DB2 Connect Server"]["Product identifier"] == "db2consv"
    assert parser_result["DB2 Connect Server"]["Version information"] == "9.7"
    assert parser_result["DB2 Connect Server"]["Enforcement policy"] == "Soft Stop"
    assert parser_result["DB2 Connect Server"]["Concurrent connect user policy"] == "Disabled"


def test_invalid_command_output():
    with pytest.raises(ParseException) as e:
        DB2Info(context_wrap(INVALID_OUTPUT))
    assert "Unable to parse db2licm info: []" == str(e.value)


def test_db2licm_doc_examples():
    env = {
        'parser_result': DB2Info(
            context_wrap(VALID_OUTPUT_MULTIPLE)),
    }
    failed, total = doctest.testmod(db2licm, globs=env)
    assert failed == 0



import logging
import os
from itertools import product

from insights.core import archives
from insights.core.context import (ClusterArchiveContext,
                                   JDRContext,
                                   HostArchiveContext,
                                   SosArchiveContext,
                                   SerializedArchiveContext)

log = logging.getLogger(__name__)


def get_all_files(path):
    all_files = []
    for f in archives.get_all_files(path):
        if os.path.isfile(f) and not os.path.islink(f):
            all_files.append(f)
    return all_files


def identify(files):
    markers = {"insights_archive.txt": SerializedArchiveContext,
               "insights_commands": HostArchiveContext,
               "sos_commands": SosArchiveContext,
               "JBOSS_HOME": JDRContext}

    for f, m in product(files, markers):
        if m in f:
            i = f.find(m)
            common_path = os.path.dirname(f[:i])
            ctx = markers[m]
            return common_path, ctx

    common_path = os.path.dirname(os.path.commonprefix(files))
    if not common_path:
        raise archives.InvalidArchive("Unable to determine common path")

    return common_path, HostArchiveContext


def create_context(path, context=None):
    top = os.listdir(path)
    arc = [os.path.join(path, f) for f in top if f.endswith(archives.COMPRESSION_TYPES)]
    if arc:
        return ClusterArchiveContext(path, all_files=arc)

    all_files = get_all_files(path)
    if not all_files:
        raise archives.InvalidArchive("No files in archive")

    common_path, ctx = identify(all_files)
    context = context or ctx
    return context(common_path, all_files=all_files)



from insights.core import Syslog
from insights.tests import context_wrap

MSGINFO = """
May 18 15:13:34 lxc-rhel68-sat56 jabberd/sm[11057]: session started: jid=rhn-dispatcher-sat@lxc-rhel6-sat56.redhat.com/superclient
May 18 15:13:36 lxc-rhel68-sat56 wrapper[11375]: --> Wrapper Started as Daemon
May 18 15:13:36 lxc-rhel68-sat56 wrapper[11375]: Launching a JVM...
May 18 15:24:28 lxc-rhel68-sat56 yum[11597]: Installed: lynx-2.8.6-27.el6.x86_64
May 18 15:36:19 lxc-rhel68-sat56 yum[11954]: Updated: sos-3.2-40.el6.noarch
Apr 22 10:35:01 boy-bona CROND[27921]: (root) CMD (/usr/lib64/sa/sa1 -S DISK 1 1)
Apr 22 10:37:32 boy-bona crontab[28951]: (root) LIST (root)
Apr 22 10:40:01 boy-bona CROND[30677]: (root) CMD (/usr/lib64/sa/sa1 -S DISK 1 1)
Apr 22 10:41:13 boy-bona crontab[32515]: (root) LIST (root)
Apr 29 11:33:36 kvmr7u5 ehtest: crontab[12345]: {
April 29 11:33:36 kvmr7u5 ehtest: crontab[12345]: { # this line will be skipped by `_parse_line`
May  5 03:50:01 kvmr7u5 systemd: Removed slice user-0.slice.
""".strip()


def test_syslog():
    msg_info = Syslog(context_wrap(MSGINFO))
    bona_list = msg_info.get('(root) LIST (root)')
    assert 2 == len(bona_list)
    assert bona_list[0].get('timestamp') == "Apr 22 10:37:32"
    assert bona_list[1].get('timestamp') == "Apr 22 10:41:13"
    crond = msg_info.get('CROND')
    assert 2 == len(crond)
    assert crond[0].get('procname') == "CROND[27921]"
    assert msg_info.get('jabberd/sm[11057]')[0].get('hostname') == "lxc-rhel68-sat56"
    assert msg_info.get('Wrapper')[0].get('message') == "--> Wrapper Started as Daemon"
    assert msg_info.get('Launching')[0].get('raw_message') == "May 18 15:13:36 lxc-rhel68-sat56 wrapper[11375]: Launching a JVM..."
    assert 2 == len(msg_info.get('yum'))
    crontab_logs = list(msg_info.get_logs_by_procname('crontab'))
    assert len(crontab_logs) == 2
    assert crontab_logs[1]['raw_message'] == "Apr 22 10:41:13 boy-bona crontab[32515]: (root) LIST (root)"
    systemd_logs = msg_info.get_logs_by_procname('systemd')
    assert len(list(systemd_logs)) == 1



"""
Cloud Provider
===============

Combiner for Cloud information. It uses the results of
:class:`InstalledRpms`,
:class:`YumRepoList` and
:class:`DMIDecode` parsers
The combiner uses these parsers determine the Cloud Provider.

Examples:
    >>> cp.cp_bios_vendor
    {'aws': 'amazon', 'google': '', 'azure': ''}
    >>> cp.cp_bios_version
    {'aws': '4.2.amazon', 'google': '', 'azure': ''}
    >>> cp.cp_rpms
    {'aws': ['rh-amazon-rhui-client-2.2.124-1.el7'], 'google': [], 'azure': []}
    >>> cp.cp_yum
    {'aws': [], 'google': [], 'azure': ['rhui-microsoft-azure-rhel7-2.2-74']}
    >>> cp.cp_asset_tag
    '7783-7084-3265-9085-8269-3286-77'
    >>> cp.cp_uuid
    'EC2F58AF-2DAD-C57E-88C0-A81CB6084290'

"""

from insights.core.plugins import combiner
from insights.parsers.installed_rpms import InstalledRpms
from insights.parsers.dmidecode import DMIDecode
from insights.parsers.yum import YumRepoList
from collections import namedtuple


@combiner([InstalledRpms, DMIDecode, YumRepoList])
class CloudProvider(object):
    """
    Combiner class to provide cloud vendor facts

    Attributes:
        cp_bios_vendor (dict): Dictionary containing a value , for each provider,
            of Bios vendor used to determine cloud provider. Each providers value will be
            empty if none found
        cp_bios_version (dict): Dictionary containing a value, for each provider,
            of Bios version used to determine cloud provider. Each providers value will be
            empty if none found
        cp_rpms (dict): Dictionary containing a list, for each provider, of rpm information
            used to determine cloud provider. Each providers list will be empty if no matches
            found
        cp_yum (dict): Dictionary containing a list, for each provider, of yum repo information
            used to determine cloud provider. Each providers list will be empty if no matches
            found
        asset_tag (dict) Dictionary containing a value, for each provider, of rpm information
            used to determine cloud provider. Each providers value will be empty if no matches
            found
        uuid (dict) Dictionary containing a value, for each provider, of rpm information
            used to determine cloud provider. Each providers value will be empty if no matches
            found
    """

    __CP = namedtuple('CP', 'name rpm yum vv')

    __GOOGLE = __CP(name='google', rpm='google-rhui-client', yum='', vv='google')
    __AWS = __CP(name='aws', rpm='rh-amazon-rhui-client', yum='', vv='amazon')
    __AZURE = __CP(name='azure', rpm='walinuxagent', yum='rhui-microsoft-azure', vv='')
    __PROVIDERS = [__GOOGLE, __AWS, __AZURE]

    AWS = __AWS.name
    """AWS Cloud Provider Constant"""

    AZURE = __AZURE.name
    """AZURE Cloud Provider Constant"""

    GOOGLE = __GOOGLE.name
    """GOOGLE Cloud Provider Constant"""

    def __init__(self, rpms, dmidcd, yrl):

        self.cp_bios_vendor = self._get_cp_bios_vendor(dmidcd)
        self.cp_bios_version = self._get_cp_bios_version(dmidcd)
        self.cp_rpms = self._get_rpm_cp_info(rpms)
        self.cp_yum = self._get_cp_from_yum(yrl)
        self.cp_asset_tag = self._get_cp_from_asset_tag(dmidcd)
        self.cp_uuid = self._get_cp_from_uuid(dmidcd)
        self.cloud_provider = self._select_provider()

    def _select_provider(self):

        if any(value for value in self.cp_bios_vendor.values()):
            return self.__AWS.name if (self.cp_bios_vendor['aws'] and
                                      self.__AWS.vv in self.cp_bios_vendor['aws'].lower())  \
                else self.__GOOGLE.name if (self.cp_bios_vendor['google'] and
                                           self.__GOOGLE.vv in self.cp_bios_vendor['google'].lower()) \
                else self.__AZURE.name if (self.cp_bios_vendor['azure'] and self.__AZURE.vv in
                                          self.cp_bios_vendor['azure'].lower()) else None

        if any(value for value in self.cp_bios_version.values()):
            return self.__AWS.name if (self.cp_bios_version['aws'] and
                                      self.__AWS.vv in self.cp_bios_version['aws'].lower())  \
                else self.__GOOGLE.name if (self.cp_bios_version['google'] and
                                           self.__GOOGLE.vv in self.cp_bios_version['google'].lower()) \
                else self.__AZURE.name if (self.cp_bios_version['azure'] and
                                          self.__AZURE.vv in self.cp_bios_version['azure'].lower()) else None

        if any(value for value in self.cp_rpms.values()):
            return self.__AWS.name if self.cp_rpms['aws'] \
                else self.__GOOGLE.name if self.cp_rpms['google'] \
                else self.__AZURE.name if self.cp_rpms['azure'] else ''

        if self.cp_yum['azure']:
            return self.__AZURE.name

        if self.cp_asset_tag['azure']:
            return self.__AZURE.name

        if self.cp_uuid['aws']:
            return self.__AWS.name

    def _get_rpm_cp_info(self, rpms):

        prov = {'aws': [], 'google': [], 'azure': []}

        if rpms:
            for p in self.__PROVIDERS:
                for key, val in rpms.packages.items():
                    for v in val:
                        prov[p.name].append(v.package) if p.rpm in v.package.lower() else prov

        return prov

    def _get_cp_from_yum(self, yrl):

        prov = {'aws': [], 'google': [], 'azure': []}

        if yrl and hasattr(yrl, 'data'):
            for p in self.__PROVIDERS:
                for yval in yrl.data:
                    prov[p.name].append(yval.get('id').lower()) \
                        if p.yum and p.yum in yval.get('id').lower() \
                        else prov

        return prov

    def _get_cp_from_asset_tag(self, dmidcd):

        prov = {'aws': '', 'google': '', 'azure': ''}

        if dmidcd and hasattr(dmidcd, 'data'):
            ch_info = dmidcd.data.get('chassis_information')
            if ch_info:
                asset_tag = ch_info[0].get('asset_tag')
                prov['azure'] = asset_tag if asset_tag == '7783-7084-3265-9085-8269-3286-77' else ''
        return prov

    def _get_cp_bios_vendor(self, dmidcd):

        prov = {'aws': '', 'google': '', 'azure': ''}

        if dmidcd and dmidcd.bios:
            for p in self.__PROVIDERS:
                prov[p.name] = dmidcd.bios.get('vendor') if p.vv and p.vv in dmidcd.bios.get('vendor').lower() \
                    else ''
        return prov

    def _get_cp_bios_version(self, dmidcd):

        prov = {'aws': '', 'google': '', 'azure': ''}

        if dmidcd and dmidcd.bios:
            for p in self.__PROVIDERS:
                prov[p.name] = dmidcd.bios.get('version') if p.vv and p.vv in dmidcd.bios.get('version').lower() \
                    else ''
        return prov

    def _get_cp_from_uuid(self, dmidcd):

        prov = {'aws': '', 'google': '', 'azure': ''}

        if dmidcd and dmidcd.bios:
            prov['aws'] = dmidcd.system_info.get('uuid') if dmidcd.system_info.get('uuid').lower().startswith('ec2') \
                else ''
        return prov



from insights.parsers import scheduler
from insights.tests import context_wrap

SDA_SCHEDULER = '''
noop deadline [cfq]
'''.strip()

SDB_SCHEDULER = '''
noop [deadline] cfq
'''.strip()

VDC_SCHEDULER = '''
[noop] deadline cfq
'''.strip()

SDA_PATH = "/sys/block/sda/queue/scheduler"
SDB_PATH = "/sys/block/sdb/queue/scheduler"
VDC_PATH = "/sys/block/vdc/queue/scheduler"


def test_scheduler_cfq():
    r = scheduler.Scheduler(context_wrap(SDA_SCHEDULER, SDA_PATH))
    assert r.data["sda"] == '[cfq]'


def test_scheduler_deadline():
    r = scheduler.Scheduler(context_wrap(SDB_SCHEDULER, SDB_PATH))
    assert r.data["sdb"] == '[deadline]'


def test_scheduler_noop():
    r = scheduler.Scheduler(context_wrap(VDC_SCHEDULER, VDC_PATH))
    assert r.data["vdc"] == '[noop]'



from functools import partial
from insights.specs import Specs
from insights.core.context import SosArchiveContext
from insights.core.spec_factory import simple_file, first_of, first_file, glob_file

first_file = partial(first_file, context=SosArchiveContext)
glob_file = partial(glob_file, context=SosArchiveContext)
simple_file = partial(simple_file, context=SosArchiveContext)


class SosSpecs(Specs):
    auditctl_status = simple_file("sos_commands/auditd/auditctl_-s")
    blkid = simple_file("sos_commands/block/blkid_-c_.dev.null")
    candlepin_log = first_of([
        simple_file("/var/log/candlepin/candlepin.log"),
        simple_file("sos_commands/foreman/foreman-debug/var/log/candlepin/candlepin.log")
    ])
    candlepin_error_log = first_of([
        simple_file("var/log/candlepin/error.log"),
        simple_file(r"sos_commands/foreman/foreman-debug/var/log/candlepin/error.log")
    ])
    catalina_out = glob_file("var/log/tomcat*/catalina.out")
    catalina_server_log = glob_file("var/log/tomcat*/catalina*.log")
    ceph_osd_tree_text = simple_file("sos_commands/ceph/ceph_osd_tree")
    ceph_report = simple_file("sos_commands/ceph/ceph_report")
    ceph_health_detail = simple_file("sos_commands/ceph/ceph_health_detail_--format_json-pretty")
    chkconfig = first_file(["sos_commands/startup/chkconfig_--list", "sos_commands/services/chkconfig_--list"])
    date = first_of([simple_file("sos_commands/general/date"), simple_file("sos_commands/date/date")])
    df__al = first_file(["sos_commands/filesys/df_-al", "sos_commands/filesys/df_-al_-x_autofs"])
    display_java = simple_file("sos_commands/java/alternatives_--display_java")
    docker_info = simple_file("sos_commands/docker/docker_info")
    docker_list_containers = first_file(["sos_commands/docker/docker_ps_-a", "sos_commands/docker/docker_ps"])
    docker_list_images = simple_file("sos_commands/docker/docker_images")
    docker_image_inspect = glob_file("sos_commands/docker/docker_inspect_*")
    dmesg = first_file(["sos_commands/kernel/dmesg", "var/log/dmesg"])
    dmidecode = simple_file("sos_commands/hardware/dmidecode")
    dmsetup_info = simple_file("sos_commands/devicemapper/dmsetup_info_-c")
    dumpe2fs_h = glob_file("sos_commands/filesys/dumpe2fs_-h_*")
    ethtool = glob_file("sos_commands/networking/ethtool_*", ignore="ethtool_-.*")
    ethtool_S = glob_file("sos_commands/networking/ethtool_-S_*")
    ethtool_T = glob_file("sos_commands/networking/ethtool_-T_*")
    ethtool_a = glob_file("sos_commands/networking/ethtool_-a_*")
    ethtool_c = glob_file("sos_commands/networking/ethtool_-c_*")
    ethtool_g = glob_file("sos_commands/networking/ethtool_-g_*")
    ethtool_i = glob_file("sos_commands/networking/ethtool_-i_*")
    ethtool_k = glob_file("sos_commands/networking/ethtool_-k_*")
    fdisk_l_sos = first_of([glob_file(r"sos_commands/filesys/fdisk_-l_*"), glob_file(r"sos_commands/block/fdisk_-l_*")])
    foreman_production_log = first_of([simple_file("/var/log/foreman/production.log"), simple_file("sos_commands/foreman/foreman-debug/var/log/foreman/production.log")])
    foreman_proxy_conf = first_of([simple_file("/etc/foreman-proxy/settings.yml"), simple_file("sos_commands/foreman/foreman-debug/etc/foreman-proxy/settings.yml")])
    foreman_proxy_log = first_of([simple_file("/var/log/foreman-proxy/proxy.log"), simple_file("sos_commands/foreman/foreman-debug/var/log/foreman-proxy/proxy.log")])
    foreman_satellite_log = first_of([simple_file("/var/log/foreman-installer/satellite.log"), simple_file("sos_commands/foreman/foreman-debug/var/log/foreman-installer/satellite.log")])
    foreman_ssl_access_ssl_log = first_file(["var/log/httpd/foreman-ssl_access_ssl.log", r"sos_commands/foreman/foreman-debug/var/log/httpd/foreman-ssl_access_ssl.log"])
    getcert_list = first_file(["sos_commands/ipa/ipa-getcert_list", "sos_commands/ipa/getcert_list"])
    gluster_v_info = simple_file("sos_commands/gluster/gluster_volume_info")
    gluster_v_status = simple_file("sos_commands/gluster/gluster_volume_status")
    gluster_peer_status = simple_file("sos_commands/gluster/gluster_peer_status")
    hostname = first_file(["sos_commands/general/hostname_-f", "sos_commands/host/hostname_-f"])
    hostname_default = first_file(["sos_commands/general/hostname", "sos_commands/host/hostname", "/etc/hostname", "hostname"])
    hostname_short = first_file(["sos_commands/general/hostname", "sos_commands/host/hostname", "/etc/hostname", "hostname"])
    installed_rpms = simple_file("installed-rpms")
    ip_addr = first_of([simple_file("sos_commands/networking/ip_-d_address"), simple_file("sos_commands/networking/ip_address")])
    ip_route_show_table_all = simple_file("sos_commands/networking/ip_route_show_table_all")
    ip_s_link = first_of([simple_file("sos_commands/networking/ip_-s_-d_link"), simple_file("sos_commands/networking/ip_-s_link"), simple_file("sos_commands/networking/ip_link")])
    iptables = first_file(["/etc/sysconfig/iptables", "/etc/sysconfig/iptables.save"])
    journal_since_boot = first_of([simple_file("sos_commands/logs/journalctl_--no-pager_--boot"), simple_file("sos_commands/logs/journalctl_--no-pager_--catalog_--boot")])
    locale = simple_file("sos_commands/i18n/locale")
    lsblk = simple_file("sos_commands/block/lsblk")
    lscpu = simple_file("sos_commands/processor/lscpu")
    lsinitrd = simple_file("sos_commands/boot/lsinitrd")
    lsof = simple_file("sos_commands/process/lsof_-b_M_-n_-l")
    lsmod = simple_file("sos_commands/kernel/lsmod")
    lspci = first_of([
        simple_file("sos_commands/pci/lspci_-nnvv"),
        simple_file("sos_commands/pci/lspci_-nvv"),
        simple_file("sos_commands/pci/lspci")
    ])
    lsscsi = simple_file("sos_commands/scsi/lsscsi")
    ls_dev = first_file(["sos_commands/block/ls_-lanR_.dev", "sos_commands/devicemapper/ls_-lanR_.dev"])
    lvs = first_file(["sos_commands/lvm2/lvs_-a_-o_lv_tags_devices_--config_global_locking_type_0", "sos_commands/lvm2/lvs_-a_-o_devices"])
    mount = simple_file("sos_commands/filesys/mount_-l")
    multipath__v4__ll = first_file(["sos_commands/multipath/multipath_-v4_-ll", "sos_commands/devicemapper/multipath_-v4_-ll"])
    netstat = first_file(["sos_commands/networking/netstat_-neopa", "sos_commands/networking/netstat_-W_-neopa", "sos_commands/networking/netstat_-T_-neopa"])
    netstat_agn = first_of([simple_file("sos_commands/networking/netstat_-agn"), simple_file("sos_commands/networking/netstat_-W_-agn"), simple_file("sos_commands/networking/netstat_-T_-agn")])
    netstat_s = simple_file("sos_commands/networking/netstat_-s")
    nmcli_dev_show = simple_file("sos_commands/networking/nmcli_device_show")
    nmcli_dev_show_sos = glob_file(["sos_commands/networking/nmcli_dev_show_*", "sos_commands/networkmanager/nmcli_dev_show_*"])
    ntptime = simple_file("sos_commands/ntp/ntptime")
    pcs_config = simple_file("sos_commands/pacemaker/pcs_config")
    pcs_quorum_status = simple_file("sos_commands/pacemaker/pcs_quorum_status")
    pcs_status = simple_file("sos_commands/pacemaker/pcs_status")
    ps_alxwww = simple_file("sos_commands/process/ps_alxwww")
    ps_aux = first_file(["sos_commands/process/ps_aux", "sos_commands/process/ps_auxwww", "sos_commands/process/ps_auxcww"])
    ps_auxcww = first_file(["sos_commands/process/ps_auxcww", "sos_commands/process/ps_auxwww", "sos_commands/process/ps_aux"])
    ps_auxww = first_file(["sos_commands/process/ps_auxww", "sos_commands/process/ps_auxwww", "sos_commands/process/ps_aux", "sos_commands/process/ps_auxcww"])
    puppet_ssl_cert_ca_pem = simple_file("sos_commands/foreman/foreman-debug/var/lib/puppet/ssl/certs/ca.pem")
    pvs = first_file(["sos_commands/lvm2/pvs_-a_-v_-o_pv_mda_free_pv_mda_size_pv_mda_count_pv_mda_used_count_pe_start_--config_global_locking_type_0", "sos_commands/lvm2/pvs_-a_-v"])
    qpid_stat_q = first_of([
        simple_file("qpid_stat_queues"),
        simple_file("qpid-stat-q"),
        simple_file("sos_commands/foreman/foreman-debug/qpid_stat_queues"),
        simple_file("sos_commands/foreman/foreman-debug/qpid-stat-q")
    ])
    qpid_stat_u = first_of([
        simple_file("qpid_stat_subscriptions"),
        simple_file("qpid-stat-u"),
        simple_file("sos_commands/foreman/foreman-debug/qpid_stat_subscriptions"),
        simple_file("sos_commands/foreman/foreman-debug/qpid-stat-u")
    ])
    rabbitmq_report = simple_file("sos_commands/rabbitmq/rabbitmqctl_report")
    rabbitmq_report_of_containers = glob_file("sos_commands/rabbitmq/docker_exec_-t_rabbitmq-bundle-docker-*_rabbitmqctl_report")
    rhn_charsets = first_file(["sos_commands/satellite/rhn-charsets", "sos_commands/rhn/rhn-charsets"])
    root_crontab = first_file(["sos_commands/crontab/root_crontab", "sos_commands/cron/root_crontab"])
    route = simple_file("sos_commands/networking/route_-n")
    sestatus = simple_file("sos_commands/selinux/sestatus_-b")
    subscription_manager_list_consumed = first_file([
        'sos_commands/yum/subscription-manager_list_--consumed',
        'sos_commands/subscription_manager/subscription-manager_list_--consumed',
        'sos_commands/general/subscription-manager_list_--consumed']
    )
    subscription_manager_list_installed = first_file([
        'sos_commands/yum/subscription-manager_list_--installed',
        'sos_commands/subscription_manager/subscription-manager_list_--installed',
        'sos_commands/general/subscription-manager_list_--installed']
    )
    sysctl = simple_file("sos_commands/kernel/sysctl_-a")
    systemctl_list_unit_files = simple_file("sos_commands/systemd/systemctl_list-unit-files")
    systemctl_list_units = first_file(["sos_commands/systemd/systemctl_list-units", "sos_commands/systemd/systemctl_list-units_--all"])
    teamdctl_config_dump = glob_file("sos_commands/teamd/teamdctl_*_config_dump")
    teamdctl_state_dump = glob_file("sos_commands/teamd/teamdctl_*_state_dump")
    uname = simple_file("sos_commands/kernel/uname_-a")
    uptime = first_of([simple_file("sos_commands/general/uptime"), simple_file("sos_commands/host/uptime")])
    var_qemu_xml = glob_file(r"var/run/libvirt/qemu/*.xml")
    vdsm_import_log = glob_file("var/log/vdsm/import/import-*.log")
    vgdisplay = first_file(["sos_commands/lvm2/vgdisplay_-vv_--config_global_locking_type_0", "sos_commands/lvm2/vgdisplay_-vv"])
    vgs = first_file(["sos_commands/lvm2/vgs_-v_-o_vg_mda_count_vg_mda_free_vg_mda_size_vg_mda_used_count_vg_tags_--config_global_locking_type_0", "sos_commands/lvm2/vgs_-v"])
    xfs_info = glob_file("sos_commands/xfs/xfs_info*")
    yum_repolist = simple_file("sos_commands/yum/yum_-C_repolist")



"""
PostgreSQLLog - file ``/var/lib/pgsql/data/pg_log/postgresql-*.log``
====================================================================
"""

from .. import LogFileOutput, parser
from insights.specs import Specs


@parser(Specs.postgresql_log)
class PostgreSQLLog(LogFileOutput):
    """
    Read the PostgreSQL log files.  Uses the ``LogFileOutput`` class parser
    functionality.

    .. note::
        Please refer to its super-class :class:`insights.core.LogFileOutput`


    The PostgreSQL log files contain no dates or times by default::

        LOG:  shutting down
        LOG:  database system is shut down
        LOG:  database system was shut down at 2015-03-31 05:05:12 UTC
        LOG:  database system is ready to accept connections
        LOG:  autovacuum launcher started

    Because this parser reads multiple log files, the contents of the shared
    parser information are a list of parsed files.  This means at present
    that you will need to iterate through the log file objects in that list
    in order to find what you want.

    Examples:
        >>> for log in shared(PacemakerLog):
        ...     print "File:", logs.file_path
        ...     print "Startups:", len(logs.get('ready to accept connections'))
        ...
        File: /var/log/pgsql/data/pg_log/postgresql-Fri
        Startups: 1
        File: /var/log/pgsql/data/pg_log/postgresql-Mon
        Startups: 0
        File: /var/log/pgsql/data/pg_log/postgresql-Sat
        Startups: 0
        File: /var/log/pgsql/data/pg_log/postgresql-Sun
        Startups: 2
        File: /var/log/pgsql/data/pg_log/postgresql-Thu
        Startups: 0
        File: /var/log/pgsql/data/pg_log/postgresql-Tue
        Startups: 0
        File: /var/log/pgsql/data/pg_log/postgresql-Wed
        Startups: 0
    """
    pass



from __future__ import print_function
import inspect
import sys
from datetime import datetime

from jinja2 import Template

from insights import dr, rule
from insights.core.context import ExecutionContext
from insights.core.spec_factory import ContentProvider
from insights.core.plugins import is_datasource
from insights.formats import render, Formatter


class TemplateFormat(Formatter):
    """
    Subclasses should implement create_template_context to return a dictionary
    to use when rendering the jinja2 template defined by the class level
    TEMPLATE attribute.
    """

    TEMPLATE = ""
    """ jinja2 template to use for rule result rendering. """

    def find_root(self):
        """
        Finds the root directory used during the evaluation. Note this could be
        a non-existent temporary directory if analyzing an archive.
        """
        for comp in self.broker:
            try:
                if issubclass(comp, ExecutionContext):
                    return self.broker[comp].root
            except:
                pass
        return "Unknown"

    def get_datasources(self, comp, broker):
        """
        Get the most relevant activated datasources for each rule.
        """
        graph = dr.get_dependency_graph(comp)
        ds = []
        for cand in graph:
            if cand in broker and is_datasource(cand):
                val = broker[cand]
                if not isinstance(val, list):
                    val = [val]

                results = []
                for v in val:
                    if isinstance(v, ContentProvider):
                        results.append(v.cmd or v.path or "python implementation")
                ds.extend(results)
        return ds

    def collect_rules(self, comp, broker):
        """
        Rule results are stored as dictionaries in the ``self.rules`` list.
        Each dictionary contains the folowing keys:
            name: fully qualified name of the rule
            id: fully qualified rule name with "." replaced with "_"
            response_type: the class name of the response object (make_info, etc.)
            body: rendered content for the rule as provided in the rule module.
            mod_doc: pydoc of the module containing the rule
            rule_doc: pydoc of the rule
            rule_path: absolute path to the source file of the rule
            datasources: sorted list of command or files contributing to the rule
        """
        if comp in broker:
            name = dr.get_name(comp)
            rule_id = name.replace(".", "_")
            val = broker[comp]
            self.rules.append({
                "name": name,
                "id": rule_id,
                "response_type": type(val).__name__,
                "body": render(comp, val),
                "mod_doc": sys.modules[comp.__module__].__doc__ or "",
                "rule_doc": comp.__doc__ or "",
                "source_path": inspect.getabsfile(comp),
                "datasources": sorted(set(self.get_datasources(comp, broker)))
            })

    def preprocess(self):
        """
        Watches rules go by as they evaluate and collects information about
        them for later display in postprocess.
        """
        self.start_time = datetime.utcnow().strftime("%Y-%m-%d %H:%M:%S")
        self.rules = []
        self.broker.add_observer(self.collect_rules, rule)

    def create_template_context(self):
        raise NotImplementedError()

    def postprocess(self):
        """
        Builds a dictionary of rule data as context for a jinja2 template that
        renders the final output.
        """
        ctx = self.create_template_context()
        print(Template(self.TEMPLATE).render(ctx), file=self.stream)



"""
Services - check ChkConfig and systemd UnitFiles
================================================

This combiner provides information about whether a given service is
enabled, using parsers for ``chkconfig`` (for RHEL 5 and 6) and ``systemd
list-unit-files`` (for RHEL 7 and above).

Examples:

    >>> svcs = shared[Services]
    >>> svcs.is_on('atd') # Can be 'atd' or 'atd.service'.
    True
    >>> svcs.is_on('systemd-journald.service')
    True
    >>> 'atd' in svcs
    True
    >>> svcs.service_line('atd')
    'atd.service                                 enabled'
    >>> 'disabled_service' in svcs
    False
    >>> 'nonexistent_service' in svcs
    False
"""

from ..core.plugins import combiner
from ..parsers import chkconfig
from ..parsers.systemd import unitfiles


@combiner([chkconfig.ChkConfig, unitfiles.UnitFiles])
class Services(object):
    """
    A combiner for working with enabled services, independent of which
    version of RHEL is in use.

    The interface closely follows the models of ChkConfig and UnitFiles:

    * ``is_on(service_name)`` and the ``service_name in Services``
      method return whether the service given is present **and enabled**.
    * ``service_line(service_name)`` returns the actual line that contained
      the service name.
    """
    def __init__(self, chk_config, unit_files):
        self.services = {}
        self.parsed_lines = {}
        # PJW 2017-02-28 - It seems completely bizarre that a system would
        # have systemd installed and also have chkconfig output.  But I'm
        # leaving these as updates instead of equals just in case...
        if chk_config:
            self.services.update(chk_config.services)
            self.parsed_lines.update(chk_config.parsed_lines)
        if unit_files:
            self.services.update(unit_files.services)
            self.parsed_lines.update(unit_files.parsed_lines)

    def is_on(self, service_name):
        """
        Checks if the service is enabled on the system.

        Args:
            service_name (str): service name (with or without '.service'
                suffix).

        Returns:
            bool: True if service is enabled, False otherwise.
        """
        return self.services.get(service_name + '.service',
                                 self.services.get(service_name, False))

    def service_line(self, service_name):
        """
        Returns the relevant line from the service listing.

        Args:
            service_name (str): service name (with or without '.service'
                suffix).

        Returns:
            str: True if service is enabled, False otherwise.
        """
        return self.parsed_lines.get(service_name + '.service',
                                     self.parsed_lines.get(service_name, ''))

    def __contains__(self, service_name):
        """
        Checks if the service **is enabled** in systemctl.  If you want
        to check whether a service is present but disabled, use the
        ``service_line`` method.

        Args:
            service_name (str): service name including '.service'

        Returns:
            bool: True if service is enabled, False otherwise
        """
        return self.is_on(service_name)



"""
HaproxyCfg - file ``/etc/haproxy/haproxy.cfg``
==============================================

Contents of the `haproxy.cfg` file look like::

    global
        daemon
        group       haproxy
        log         /dev/log local0
        user        haproxy
        maxconn     20480
        pidfile     /var/run/haproxy.pid

    defaults
        retries     3
        maxconn     4096
        log         global
        timeout     http-request 10s
        timeout     queue 1m
        timeout     connect 10s

If there are duplicate key items, merge them in to one. Like::

    option  tcpka
                            }--->    option: ["tcpka","tcplog"]
    option  tcplog

Examples:
    >>> cfg = shared[HaproxyCfg]
    >>> cfg.data['global']
    {"daemon": "", "group": "haproxy", "log": " /dev/log local0",
     "user": "haproxy", "maxconn": "20480", "pidfile": "/var/run/haproxy.pid"}
    >>> cfg.data['global']['group']
    "haproxy"
    >>> 'global' in cfg.data
    True
    >>> 'user' in cfg.data.get('global')
    True
    """
from .. import Parser, parser
from insights.specs import Specs


def _parse_content(content):
    SECTION_NAMES = ("global", "defaults", "frontend", "backend", "listen")
    haproxy_dict = {}
    section_dict = {}
    for line in content:
        line = line.strip()
        if line.startswith("#") or line == "":
            continue
        values = line.split(None, 1)
        if values[0] in SECTION_NAMES:
            # new section like  global:{} or listen mysql: {}
            section_dict = {}
            i_key = values[0] if len(values) == 1 else values[0] + " " + values[1]
            haproxy_dict.update({i_key: section_dict})
        else:
            # handle attributes in one section
            if len(values) == 1:
                section_dict[line] = ""
            else:
                attr_key = values[0]
                attr_value = values[1]
                if attr_key in section_dict:
                    # if it is not list, convert it to list
                    if not isinstance(section_dict[attr_key], list):
                        section_dict[attr_key] = [section_dict[attr_key]]
                    section_dict[attr_key].append(attr_value)
                else:
                    section_dict[attr_key] = attr_value
    return haproxy_dict


@parser(Specs.haproxy_cfg)
class HaproxyCfg(Parser):
    """Class to parse file ``haproxy.cfg``."""
    def parse_content(self, content):
        self.data = _parse_content(content)



import doctest
from insights.tests import context_wrap
from insights.parsers import mlx4_port
from insights.parsers.mlx4_port import Mlx4Port

MLX4_CONTENT = """
ib
"""

MLX4_CONTENT_ML = """  line 1
  line 2"""

MLX4_PATH = '/sys/bus/pci/devices/0000:0c:00.0/mlx4_port1'


def test_mlx4_port():
    result = Mlx4Port(context_wrap(MLX4_CONTENT, path=MLX4_PATH))
    assert result.name == 'mlx4_port1'
    assert result.contents == ['ib']
    result = Mlx4Port(context_wrap(MLX4_CONTENT_ML, path=MLX4_PATH))
    assert result.name == 'mlx4_port1'
    assert result.contents == ['line 1', 'line 2']


def test_mlx4_docs():
    env = {'mlx4_port': Mlx4Port(context_wrap(MLX4_CONTENT, path=MLX4_PATH))}
    failed, total = doctest.testmod(mlx4_port, globs=env)
    assert failed == 0



# from insights.parsers.virt_what import VirtWhat as VWP
# from insights.combiners.virt_what import VirtWhat
from insights.combiners.cloud_provider import CloudProvider
from insights.parsers.installed_rpms import InstalledRpms as IRPMS
from insights.parsers.dmidecode import DMIDecode
from insights.parsers.yum import YumRepoList
from insights.tests import context_wrap

RPMS = """
gnome-terminal-3.28.2-2.fc28.x86_64
python3-IPy-0.81-21.fc28.noarch
gnu-free-serif-fonts-20120503-17.fc28.noarch
""".strip()

RPMS_AWS = """
gnome-terminal-3.28.2-2.fc28.x86_64
python3-IPy-0.81-21.fc28.noarch
gnu-free-serif-fonts-20120503-17.fc28.noarch
rh-amazon-rhui-client-2.2.124-1.el7
""".strip()

RPMS_GOOGLE = """
gnome-terminal-3.28.2-2.fc28.x86_64
python3-IPy-0.81-21.fc28.noarch
gnu-free-serif-fonts-20120503-17.fc28.noarch
google-rhui-client-5.1.100-1.el7
google-rhui-client-5.1.100-1.el6
""".strip()

RPMS_AZURE = """
gnome-terminal-3.28.2-2.fc28.x86_64
python3-IPy-0.81-21.fc28.noarch
gnu-free-serif-fonts-20120503-17.fc28.noarch
WALinuxAgent-2.2.18-1.el7
""".strip()

YUM_REPOLIST_AZURE = """
Loaded plugins: enabled_repos_upload, package_upload, product-id, search-
              : disabled-repos, security, subscription-manager
repo id                                repo name                          status
rhel-6-server-rpms                     Red Hat Enterprise Linux 6 Server  20584
rhel-6-server-satellite-tools-6.3-rpms Red Hat Satellite Tools 6.3 (for R    40
rhui-microsoft-azure-rhel7-2.2-74      Red Hat Software Collect             600
repolist: 21224
Uploading Enabled Repositories Report
Loaded plugins: product-id, subscription-manager
""".strip()

YUM_REPOLIST_NOT_AZURE = """
Loaded plugins: enabled_repos_upload, package_upload, product-id, search-
              : disabled-repos, security, subscription-manager
repo id                                repo name                          status
rhel-6-server-rpms                     Red Hat Enterprise Linux 6 Server  20584
rhel-6-server-satellite-tools-6.3-rpms Red Hat Satellite Tools 6.3 (for R    40
repolist: 21224
Uploading Enabled Repositories Report
Loaded plugins: product-id, subscription-manager
""".strip()


DMIDECODE = '''
# dmidecode 2.11
SMBIOS 2.7 present.
188 structures occupying 5463 bytes.
Table at 0xBFFCB000.

Handle 0x0000, DMI type 0, 24 bytes
BIOS Information
\tVendor: HP
\tVersion: P70
\tRelease Date: 03/01/2013
\tAddress: 0xF0000
\tRuntime Size: 64 kB
\tROM Size: 8192 kB
\tCharacteristics:
\t\tPCI is supported
\t\tPNP is supported
\t\tBIOS is upgradeable
\t\tBIOS shadowing is allowed
\t\tESCD support is available
\t\tBoot from CD is supported
\t\tSelectable boot is supported
\t\tEDD is supported
\t\t5.25"/360 kB floppy services are supported (int 13h)
\t\t5.25"/1.2 MB floppy services are supported (int 13h)
\t\t3.5"/720 kB floppy services are supported (int 13h)
\t\tPrint screen service is supported (int 5h)
\t\tj042 keyboard services are supported (int 9h)
\t\tSerial services are supported (int 14h)
\t\tPrinter services are supported (int 17h)
\t\tCGA/mono video services are supported (int 10h)
\t\tACPI is supported
\t\tUSB legacy is supported
\t\tBIOS boot specification is supported
\t\tFunction key-initiated network boot is supported
\t\tTargeted content distribution is supported
\tFirmware Revision: 1.22

Handle 0x0100, DMI type 1, 27 bytes
System Information
\tManufacturer: HP
\tProduct Name: ProLiant DL380p Gen8
\tVersion: Not Specified
\tSerial Number: 2M23360006
\tUUID: 34373936-3439-4D32-3233-333630303036
\tWake-up Type: Power Switch
\tSKU Number: 697494-S01
\tFamily: ProLiant

Handle 0x0300, DMI type 3, 21 bytes
chassis information
\tmanufacturer: hp
\ttype: rack mount chassis
\tlock: not present
\tversion: not specified
\tserial number: 2m23360006
\tasset Tag:
\tBoot-up State: Safe
\tPower Supply State: Safe
\tThermal State: Safe
\tSecurity Status: Unknown
\tOEM Information: 0x00000000
\tHeight: 2 U
\tNumber Of Power Cords: 2
\tContained Elements: 0

Handle 0x0401, DMI type 4, 32 bytes
Processor Information
\tSocket Designation: CPU 1
\tType: Central Processor
\tFamily: Other
\tManufacturer: Bochs
\tID: A1 06 02 00 FD FB 8B 17
\tVersion: Not Specified
\tVoltage: Unknown
\tExternal Clock: Unknown
\tMax Speed: 2000 MHz
\tCurrent Speed: 2000 MHz
\tStatus: Populated, Enabled
\tUpgrade: Other
\tL1 Cache Handle: Not Provided
\tL2 Cache Handle: Not Provided
\tL3 Cache Handle: Not Provided

Handle 0x0402, DMI type 4, 32 bytes
Processor Information
\tSocket Designation: CPU 2
\tType: Central Processor
\tFamily: Other
\tManufacturer: Bochs
\tID: A1 06 02 00 FD FB 8B 17
\tVersion: Not Specified
\tVoltage: Unknown
\tExternal Clock: Unknown
\tMax Speed: 2000 MHz
\tCurrent Speed: 2000 MHz
\tStatus: Populated, Enabled
\tUpgrade: Other
\tL1 Cache Handle: Not Provided
\tL2 Cache Handle: Not Provided
\tL3 Cache Handle: Not Provided

Handle 0x0037, DMI type 127, 4 bytes.
End Of Table
'''

DMIDECODE_BARE_METAL = '''
# dmidecode 3.0
Getting SMBIOS data from sysfs.
SMBIOS 2.7 present.
56 structures occupying 2407 bytes.
Table at 0x000EB4C0.

Handle 0x0000, DMI type 0, 24 bytes
BIOS Information
\tVendor: American Megatrends Inc.
\tVersion: 2.0a
\tRelease Date: 06/08/2012
\tAddress: 0xF0000
\tRuntime Size: 64 kB
\tROM Size: 8192 kB
\tCharacteristics:
\t\tPCI is supported
\t\tBIOS is upgradeable
\t\tBIOS shadowing is allowed
\t\tBoot from CD is supported
\t\tSelectable boot is supported
\t\tBIOS ROM is socketed
\t\tEDD is supported
\t\t5.25"/1.2 MB floppy services are supported (int 13h)
\t\t3.5"/720 kB floppy services are supported (int 13h)
\t\t3.5"/2.88 MB floppy services are supported (int 13h)
\t\tPrint screen service is supported (int 5h)
\t\t8042 keyboard services are supported (int 9h)
\t\tSerial services are supported (int 14h)
\t\tPrinter services are supported (int 17h)
\t\tACPI is supported
\t\tUSB legacy is supported
\t\tBIOS boot specification is supported
\t\tFunction key-initiated network boot is supported
\t\tTargeted content distribution is supported
\t\tUEFI is supported
\tBIOS Revision: 2.10

Handle 0x0001, DMI type 1, 27 bytes
System Information
\tManufacturer: Supermicro
\tProduct Name: X9SCL/X9SCM
\tVersion: 0123456789
\tSerial Number: 0123456789
\tUUID: 12345678-1234-1234-1234-123456681234
\tWake-up Type: Power Switch
\tSKU Number: To be filled by O.E.M.
\tFamily: To be filled by O.E.M.

Handle 0x0002, DMI type 2, 15 bytes
Base Board Information
\tManufacturer: Supermicro
\tProduct Name: X9SCL/X9SCM
\tVersion: 0123456789
\tSerial Number: 1234567812
\tAsset Tag: To be filled by O.E.M.
\tFeatures:
\t\tBoard is a hosting board
\t\tBoard is replaceable
\tLocation In Chassis: To be filled by O.E.M.
\tChassis Handle: 0x0003
\tType: Motherboard
\tContained Object Handles: 0
'''.strip()

DMIDECODE_AWS = '''
# dmidecode 2.12-dmifs
SMBIOS 2.4 present.
11 structures occupying 310 bytes.
Table at 0x000EB01F.

Handle 0x0000, DMI type 0, 24 bytes
BIOS Information
\tVendor: Xen
\tVersion: 4.2.amazon
\tRelease Date: 12/09/2016
\tAddress: 0xE8000
\tRuntime Size: 96 kB
\tROM Size: 64 kB
\tCharacteristics:
\t\tPCI is supported
\t\tEDD is supported
\t\tTargeted content distribution is supported
\tBIOS Revision: 4.2

Handle 0x0100, DMI type 1, 27 bytes
System Information
\tManufacturer: Xen
\tProduct Name: HVM domU
\tVersion: 4.2.amazon
\tSerial Number: ec2f58af-2dad-c57e-88c0-a81cb6084290
\tUUID: EC2F58AF-2DAD-C57E-88C0-A81CB6084290
\tWake-up Type: Power Switch
\tSKU Number: Not Specified
\tFamily: Not Specified

Handle 0x0300, DMI type 3, 13 bytes
Chassis Information
\tManufacturer: Xen
\tType: Other
\tLock: Not Present
\tVersion: Not Specified
\tSerial Number: Not Specified
\tAsset Tag: Not Specified
\tBoot-up State: Safe
\tPower Supply State: Safe
\tThermal State: Safe
\tSecurity Status: Unknown
'''

DMIDECODE_AWS_UUID = '''
# dmidecode 2.12-dmifs
SMBIOS 2.4 present.
11 structures occupying 310 bytes.
Table at 0x000EB01F.

Handle 0x0000, DMI type 0, 24 bytes
BIOS Information
\tVendor: Xen
\tVersion: 4.2
\tRelease Date: 12/09/2016
\tAddress: 0xE8000
\tRuntime Size: 96 kB
\tROM Size: 64 kB
\tCharacteristics:
\t\tPCI is supported
\t\tEDD is supported
\t\tTargeted content distribution is supported
\tBIOS Revision: 4.2

Handle 0x0100, DMI type 1, 27 bytes
System Information
\tManufacturer: Xen
\tProduct Name: HVM domU
\tVersion: 4.2.amazon
\tSerial Number: ec2f58af-2dad-c57e-88c0-a81cb6084290
\tUUID: EC2F58AF-2DAD-C57E-88C0-A81CB6084290
\tWake-up Type: Power Switch
\tSKU Number: Not Specified
\tFamily: Not Specified

Handle 0x0300, DMI type 3, 13 bytes
Chassis Information
\tManufacturer: Xen
\tType: Other
\tLock: Not Present
\tVersion: Not Specified
\tSerial Number: Not Specified
\tAsset Tag: Not Specified
\tBoot-up State: Safe
\tPower Supply State: Safe
\tThermal State: Safe
\tSecurity Status: Unknown
'''

DMIDECODE_GOOGLE = '''
# dmidecode 2.12-dmifs
SMBIOS 2.4 present.
11 structures occupying 310 bytes.
Table at 0x000EB01F.

Handle 0x0000, DMI type 0, 24 bytes
BIOS Information
\tVendor: Google
\tVersion: Google
\tRelease Date: 12/09/2016
\tAddress: 0xE8000
\tRuntime Size: 96 kB
\tROM Size: 64 kB
\tCharacteristics:
\t\tPCI is supported
\t\tEDD is supported
\t\tTargeted content distribution is supported
\tBIOS Revision: 4.2

Handle 0x0100, DMI type 1, 27 bytes
System Information
\tManufacturer: Xen
\tProduct Name: HVM domU
\tVersion: 4.2.amazon
\tSerial Number: ec2f58af-2dad-c57e-88c0-a81cb6084290
\tUUID: EC2F58AF-2DAD-C57E-88C0-A81CB6084290
\tWake-up Type: Power Switch
\tSKU Number: Not Specified
\tFamily: Not Specified

Handle 0x0300, DMI type 3, 13 bytes
Chassis Information
\tManufacturer: Xen
\tType: Other
\tLock: Not Present
\tVersion: Not Specified
\tSerial Number: Not Specified
\tAsset Tag: Not Specified
\tBoot-up State: Safe
\tPower Supply State: Safe
\tThermal State: Safe
\tSecurity Status: Unknown
'''

DMIDECODE_AZURE_ASSET_TAG = """
# dmidecode 3.1
Getting SMBIOS data from sysfs.
SMBIOS 2.8 present.
10 structures occupying 511 bytes.
Table at 0x000F6050.

Handle 0x0000, DMI type 0, 24 bytes
BIOS Information
\tVendor: SeaBIOS
\tVersion: 1.11.0-2.el7
\tRelease Date: 04/01/2014
\tAddress: 0xE8000
\tRuntime Size: 96 kB
\tROM Size: 64 kB
\tCharacteristics:
\t\tBIOS characteristics not supported
\t\tTargeted content distribution is supported
\tBIOS Revision: 0.0

Handle 0x0100, DMI type 1, 27 bytes
System Information
\tManufacturer: oVirt
\tProduct Name: oVirt Node
\tVersion: 7-5.1804.4.el7.centos
\tSerial Number: 30393137-3436-584D-5136-323830304E46
\tUUID: a35ae32b-ed0a-49a4-9dbb-eecf21f88aab
\tWake-up Type: Power Switch
\tSKU Number: Not Specified
\tFamily: Red Hat Enterprise Linux

Handle 0x0300, DMI type 3, 21 bytes
Chassis Information
\tManufacturer: Red Hat
\tType: Other
\tLock: Not Present
\tVersion: RHEL 7.2.0 PC (i440FX + PIIX, 1996)
\tSerial Number: Not Specified
\tAsset Tag: 7783-7084-3265-9085-8269-3286-77
\tBoot-up State: Safe
\tPower Supply State: Safe
\tThermal State: Safe
\tSecurity Status: Unknown
\tOEM Information: 0x00000000
\tHeight: Unspecified
\tNumber Of Power Cords: Unspecified
\tContained Elements: 0
"""

DMIDECODE_FAIL = "# dmidecode 2.11\n# No SMBIOS nor DMI entry point found, sorry.\n"


def test_rpm_google():
    irpms = IRPMS(context_wrap(RPMS_GOOGLE))
    dmi = DMIDecode(context_wrap(DMIDECODE))
    yrl = YumRepoList(context_wrap(YUM_REPOLIST_NOT_AZURE))
    ret = CloudProvider(irpms, dmi, yrl)
    assert ret.cloud_provider == 'google'
    assert 'google-rhui-client-5.1.100-1.el7' in ret.cp_rpms.get('google')
    assert 'google-rhui-client-5.1.100-1.el6' in ret.cp_rpms.get('google')


def test_rpm_aws():
    irpms = IRPMS(context_wrap(RPMS_AWS))
    dmi = DMIDecode(context_wrap(DMIDECODE))
    yrl = YumRepoList(context_wrap(YUM_REPOLIST_NOT_AZURE))
    ret = CloudProvider(irpms, dmi, yrl)
    assert ret.cloud_provider == 'aws'
    assert ret.cp_rpms.get('aws')[0] == 'rh-amazon-rhui-client-2.2.124-1.el7'


def test_rpm_azure():
    irpms = IRPMS(context_wrap(RPMS_AZURE))
    dmi = DMIDecode(context_wrap(DMIDECODE_BARE_METAL))
    yrl = YumRepoList(context_wrap(YUM_REPOLIST_NOT_AZURE))
    ret = CloudProvider(irpms, dmi, yrl)
    assert ret.cloud_provider == 'azure'
    assert ret.cp_rpms.get('azure')[0] == 'WALinuxAgent-2.2.18-1.el7'


def test__yum_azure():
    irpms = IRPMS(context_wrap(RPMS))
    dmi = DMIDecode(context_wrap(DMIDECODE))
    yrl = YumRepoList(context_wrap(YUM_REPOLIST_AZURE))
    ret = CloudProvider(irpms, dmi, yrl)
    assert ret.cloud_provider == 'azure'
    assert 'rhui-microsoft-azure-rhel7-2.2-74' in ret.cp_yum.get('azure')


def test__bios_version_aws():
    irpms = IRPMS(context_wrap(RPMS))
    dmi = DMIDecode(context_wrap(DMIDECODE_AWS))
    yrl = YumRepoList(context_wrap(YUM_REPOLIST_AZURE))
    ret = CloudProvider(irpms, dmi, yrl)
    assert ret.cloud_provider == 'aws'
    assert ret.cp_bios_version['aws'] == '4.2.amazon'


def test__bios_vendor_google():
    irpms = IRPMS(context_wrap(RPMS))
    dmi = DMIDecode(context_wrap(DMIDECODE_GOOGLE))
    yrl = YumRepoList(context_wrap(YUM_REPOLIST_AZURE))
    ret = CloudProvider(irpms, dmi, yrl)
    assert ret.cloud_provider == 'google'
    assert ret.cp_bios_vendor['google'] == 'Google'


def test__asset_tag_azure():
    irpms = IRPMS(context_wrap(RPMS))
    dmi = DMIDecode(context_wrap(DMIDECODE_AZURE_ASSET_TAG))
    yrl = YumRepoList(context_wrap(YUM_REPOLIST_NOT_AZURE))
    ret = CloudProvider(irpms, dmi, yrl)
    assert ret.cloud_provider == 'azure'
    assert ret.cp_asset_tag['azure'] == '7783-7084-3265-9085-8269-3286-77'


def test__uuid():
    irpms = IRPMS(context_wrap(RPMS))
    dmi = DMIDecode(context_wrap(DMIDECODE_AWS_UUID))
    yrl = YumRepoList(context_wrap(YUM_REPOLIST_NOT_AZURE))
    ret = CloudProvider(irpms, dmi, yrl)
    assert ret.cloud_provider == 'aws'
    assert ret.cp_uuid['aws'] == 'EC2F58AF-2DAD-C57E-88C0-A81CB6084290'
